import asyncio
import httpx
from typing import List, Optional, Any, Dict
from pydantic import field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict
from cachetools import LRUCache
from ragchat.definitions import BaseNode, Relation, UrlKey
from ragchat.log import get_logger
from ragchat.utils import select_model, timeit

logger = get_logger(__name__)

class EmbeddingSettings(BaseSettings):
    base_url: Optional[str] = None
    local_hosts: Optional[List[str]] = ["localhost", "host.docker.internal"]
    port: Optional[int] = None
    api_key: Optional[str] = None
    model: Optional[str] = None
    models: Optional[List[str]] = None
    dims: int = 1536
    cache_size: int = 5461

    model_config = SettingsConfigDict(case_sensitive=False, env_prefix="EMBEDDING_")

    @field_validator("models", mode="before")
    @classmethod
    def validate_models(cls, v):
        if isinstance(v, str):
            return [m.strip() for m in v.split(',')]
        return v

    def request_dict(self):
        return self.model_dump(mode='json', include={"base_url", "api_key", "model", "dims"}, exclude_none=True)

    async def initialize(self):
        """Select the best available model and API endpoint based on settings."""
        apis = set()
        if self.base_url and self.api_key:
            apis.add(UrlKey(url=self.base_url, key=self.api_key))
        port = f":{self.port}" if self.port else ""
        for host in self.local_hosts:
            apis.add(UrlKey(url=f"http://{host}{port}/v1", key="NA"))
        
        selected_model = await select_model([self.model] if self.model else self.models, apis)
        self.base_url = selected_model.url
        self.api_key = selected_model.key
        self.model = selected_model.model
        logger.info(f"Using model {selected_model.model} from URL {selected_model.url}")

class Embedder:
    def __init__(self, settings: Optional[EmbeddingSettings] = None):
        self.settings = settings or EmbeddingSettings()
        self.cache = LRUCache(maxsize=self.settings.cache_size)

    async def initialize(self):
        await self.settings.initialize()

    @staticmethod
    async def from_openai(
        texts: List[str],
        api_key: str,
        base_url: str,
        model: str,
        dims: Optional[int] = None,
    ) -> List[List[float]]:
        """
        Get embeddings from an OpenAI-compatible API for a list of texts.
        """
        url = f"{base_url}/embeddings"
        headers = {
            "Authorization": f"Bearer {api_key}",
            "Content-Type": "application/json",
        }
        payload = {
            "input": texts,
            "model": model,
        }
        if dims:
            payload["dimensions"] = dims

        async with httpx.AsyncClient(timeout=30) as client:
            response = await client.post(url, headers=headers, json=payload)
            response.raise_for_status()
            return [item["embedding"] for item in response.json()["data"]]

    async def embed(self, texts: List[str], use_cache=True, **kwargs) -> List[List[float]]:
        """Get the embeddings for the given list of texts using the configured API, with text deduplication and caching."""
        # Preprocess texts and prepare cache keys
        clean_texts = [t.replace("\n", " ") for t in texts]
        params = {**self.settings.request_dict(), **kwargs}
        model = params["model"]
        dims = params.get("dims")

        # Deduplicate texts while preserving original order information
        unique_texts = []
        text_to_unique_idx = {}
        
        for text in clean_texts:
            if text not in text_to_unique_idx:
                text_to_unique_idx[text] = len(unique_texts)
                unique_texts.append(text)
        
        # Prepare cache keys for each unique text
        unique_cache_keys = [(t, model, dims) for t in unique_texts]

        # Check cache and collect texts to fetch
        unique_embeddings = [None] * len(unique_texts)
        uncached_indices = []
        uncached_texts = []
        
        for idx, (text, cache_key) in enumerate(zip(unique_texts, unique_cache_keys)):
            if use_cache and cache_key in self.cache:
                unique_embeddings[idx] = self.cache[cache_key]
            else:
                uncached_indices.append(idx)
                uncached_texts.append(text)

        # Fetch embeddings for uncached unique texts
        if uncached_texts:
            uncached_embeddings = await self.from_openai(
                texts=uncached_texts,
                api_key=params["api_key"],
                base_url=params["base_url"],
                model=model,
                dims=dims,
            )
            # Store in cache and fill in the results
            for i, emb in enumerate(uncached_embeddings):
                unique_idx = uncached_indices[i]
                if use_cache:
                    self.cache[unique_cache_keys[unique_idx]] = emb
                unique_embeddings[unique_idx] = emb

        # Map the unique embeddings back to the original order
        embeddings = [unique_embeddings[text_to_unique_idx[text]] for text in clean_texts]
        return embeddings

    async def embed_nodes(self, nodes: List[BaseNode], use_cache: bool = True, **kwargs) -> None:
        """Embed a list of nodes by embedding their string representations."""
        embs = await self.embed([str(n) for n in nodes], use_cache, **kwargs)
        for i, n in enumerate(nodes):
            n.embedding = embs[i]

    @timeit
    async def embed_relations(self, relations: List[Relation], use_cache: bool = True, **kwargs) -> None:
        """Embed the nodes contained within a list of relations."""
        all_nodes = []
        for r in relations:
            all_nodes.extend(r.to_list())
        await self.embed_nodes(all_nodes, use_cache, **kwargs)


import re
import hashlib
import traceback
from typing import Literal, ClassVar, List, Optional, Any, Annotated, Dict, Set, Tuple, NamedTuple, Type, TypeVar
from typing_extensions import Self
from dataclasses import dataclass, fields
from litellm import Message
from pydantic import BaseModel, create_model, model_validator, ConfigDict, Field, field_validator, ValidationError, TypeAdapter, PrivateAttr
from pydantic_core import PydanticUndefined
from enum import Enum
from logging import getLogger
logger = getLogger(__name__)



####################################
# --- Flows & General ---
####################################

class classproperty:
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, cls=None):
        if cls is None:
            cls = type(instance)
        return self.func(cls)

class Flow(Enum):
    FILE = "file"
    CHAT = "chat"
    def __str__(self):
        return self.value

class MessageClassification(Enum):
    """Enum representing the possible types of messages."""
    STATEMENT = "statement"
    QUESTION = "question"
    MIXED = "mixed"
    NONE = "none"
    def __str__(self):
        return self.value

T = TypeVar("T", bound=BaseModel)

class _mappedFields(BaseModel):
    @model_validator(mode='after')
    def at_least_one(self):
        if not self.model_fields_set:
            raise ValueError("At least one field must be set.")
        return self

    def to_base(self, base: T, exclude_none: bool = True) -> T:
        """
        Maps fields from this model instance to a provided base model instance.

        Mapping rules:
        - Fields defined in this model map to base model fields with the same name.
        - Fields in this model's model_extra map to entries in base.custom with the same key.
        - If a standard field from this model doesn't match a standard base field and base has no 'custom' attribute,
        it is ignored and a warning is logged.
        - If an extra field from this model is present and base has no 'custom' attribute,
        it is ignored and a warning is logged.
        - None values are excluded if exclude_none is True.
        """
        if base is None:
            raise TypeError("The 'base' parameter cannot be None.")

        base_keys: Set[str] = set(base.model_fields.keys())
        target_base_name: str = base.__class__.__name__

        # --- Handle base.custom initialization/validation upfront ---
        base_custom_dict: Optional[Dict[str, Any]] = None
        if hasattr(base, 'custom'):
            base_custom_value = getattr(base, 'custom')
            if isinstance(base_custom_value, dict):
                base_custom_dict = base_custom_value
            else:
                # It exists but isn't a dict (or is None). Initialize it.
                if base_custom_value is not None:
                    logger.warning(
                        f"base.custom ({type(base_custom_value)}) is not a dictionary on {target_base_name}, "
                        f"overwriting with empty dict for mapping."
                    )
                setattr(base, 'custom', {})
                # Get the reference to the newly created dictionary
                base_custom_dict = getattr(base, 'custom')
        # If base does not have 'custom', base_custom_dict remains None.
        # -------------------------------------------------------------

        # 1. Map standard fields from self to base
        for model_key in self.model_fields.keys():
            value = getattr(self, model_key)

            if exclude_none and value is None:
                continue

            if model_key in base_keys:
                # Rule: Map to base field with the same name
                setattr(base, model_key, value)
            elif base_custom_dict is not None:
                # Rule: If no standard field match, map to base.custom (if available)
                base_custom_dict[model_key] = value
            else:
                # Log warning if field cannot be mapped anywhere
                logger.warning(
                    f"Standard field '{model_key}' from {self.__class__.__name__} cannot be mapped to {target_base_name}. "
                    f"No matching standard field found and {target_base_name} does not have a 'custom' attribute."
                )

        # 2. Map extra fields from self.model_extra to base.custom
        if self.model_extra: # Check if model_extra is not None or empty
            for extra_key, extra_value in self.model_extra.items():
                if exclude_none and extra_value is None:
                    continue

                if base_custom_dict is not None:
                    # Rule: Map extra fields ONLY to base.custom (if available)
                    base_custom_dict[extra_key] = extra_value
                else:
                    # Log warning if extra field cannot be mapped
                    logger.warning(
                        f"Extra field '{extra_key}' from {self.__class__.__name__} cannot be mapped to {target_base_name}. "
                        f"{target_base_name} does not have a 'custom' attribute."
                    )

        if type(base) == BaseNode:
            base._update_hash()
        
        return base

    @classmethod
    def from_base(cls: Type[T], base: BaseModel, exclude_none: bool = True) -> T:
        """
        Maps fields from a base model instance to this model class.

        Mapping rules:
        - Base model fields map to fields in this model with the same name.
        - Entries in base.custom map to fields in this model with the same key,
        overwriting standard base fields if names collide.
        - Base model fields and base.custom entries that don't match a defined field
        in this model are mapped into this model's model_extra if allowed.
        - None values are excluded if exclude_none is True.
        """
        # 1. Get data from standard base fields (excluding 'custom' itself).
        #    model_dump handles exclude and exclude_none filtering efficiently.
        source_data = base.model_dump(exclude={'custom'}, exclude_none=exclude_none)

        # 2. Get data from base.custom, if it exists and is a dictionary.
        base_custom_dict: Dict[str, Any] = {}
        if hasattr(base, 'custom'):
            if isinstance(base.custom, dict):
                base_custom_dict = base.custom
            elif base.custom is not None:
                # Log warning only if custom exists and is not None/dict
                # Use stack level 2 to point to the caller of from_base
                logger.debug("Stack trace:\n" + "".join(traceback.format_stack()))
                logger.warning(f"base.custom ({type(base.custom)}) is not a dictionary, ignoring custom fields during mapping from base.", stacklevel=2)

        # 3. Add data from base.custom to source_data, allowing it to overwrite
        #    standard fields if keys collide (as per original logic's potential).
        if base_custom_dict:
            # Filter None from custom dict if needed before updating source_data
            filtered_custom_data = {
                k: v for k, v in base_custom_dict.items()
                if not (exclude_none and v is None)
            }
            source_data.update(filtered_custom_data) # .update() handles overwriting

        # 4. Create the instance. Pydantic's constructor automatically maps
        #    keys matching model_fields and puts the rest into model_extra
        #    if 'extra' is 'allow'. This replaces the manual init_data/extra_data logic.
        instance = cls(**source_data)

        return instance

    @classmethod
    def required_fields(cls) -> set[str]:
        return {
            field_name
            for field_name, field_info in cls.model_fields.items()
            if field_info.default is PydanticUndefined and field_info.default_factory is None
        }

####################################
# --- Metadata and Filters ---
####################################

Id = str | int

PRIMITIVE_TYPES: Tuple[type, ...] = (str, int, float, bool)

Primitive = str | int | float | bool

ALLOWED_KEY_REGEX = re.compile(r'^[a-zA-Z_][a-zA-Z0-9_]*$')

class BaseMetadata(_mappedFields):
    """
    Base class for metadata models.
    """
    model_config = ConfigDict(extra="allow")

    @model_validator(mode='after')
    def validate_extra_fields(self):
        if not self.model_extra:
            return self

        allowed_value_types = PRIMITIVE_TYPES

        for key, value in self.model_extra.items():
            if not isinstance(key, str):
                 raise ValueError(f"Custom field key must be a string, got {type(key).__name__} for key '{key}'.")

            if not ALLOWED_KEY_REGEX.fullmatch(key):
                 raise ValueError(
                     f"Custom field key '{key}' contains invalid characters or format. "
                     f"Keys must match regex '{ALLOWED_KEY_REGEX.pattern}'."
                 )

            # Existing Value Validation
            if not isinstance(value, allowed_value_types):
                raise ValueError(
                    f"Custom field '{key}' has an invalid value type. "
                    f"Expected one of {[t.__name__ for t in allowed_value_types]}, but got {type(value).__name__}."
                )

        return self
    
class ChatMetadata(BaseMetadata):
    """Metadata specific to `chat` flow."""
    _flow: Flow = PrivateAttr(Flow.CHAT)

    user_id: Id = Field(description="`Indexed` A user ID is an isolated search space.")
    chat_id: Optional[Id] = Field(None, description="`Indexed` Metadata for filtering.")
    
class FileMetadata(BaseMetadata):
    """Metadata specific to `file` flow."""
    _flow: Flow = PrivateAttr(Flow.FILE)

    path: str = Field(description="Needed to process the file.")
    folder_id: Id = Field(description="`Indexed` A folder ID is an isolated search space.")
    user_id: Optional[Id] = Field(None, description="`Indexed` Metadata for filtering.")
    file_id: Optional[Id] = Field(None, description="`Indexed` Metadata for filtering.")
    title: Optional[str] = Field(None, description="Metadata for filtering.")

Metadata = ChatMetadata | FileMetadata

class Operator(Enum):
    EQ = "="
    LT = "<"
    GT = ">"
    IN = "in"
    def __str__(self):
        return self.value

class Condition(BaseModel):
    """Represents an operator-value filter condition."""
    operator: Operator
    value: Primitive | List[Primitive]

    @model_validator(mode='after')
    def validate_operator_value_compatibility(self):
        """Validate operator/value compatibility within the Condition."""
        if self.operator == Operator.IN:
            if not isinstance(self.value, list):
                raise ValueError(
                    f"Operator '{self.operator.name}' requires a list value, "
                    f"but got {type(self.value).__name__}."
                )
        elif self.operator in (Operator.EQ, Operator.LT, Operator.GT):
            if isinstance(self.value, list):
                 raise ValueError(
                     f"Operator '{self.operator.name}' requires a single value, "
                     f"but got a list."
                 )
        return self


Filter = Primitive | List[Primitive] | List[Condition]
filter_adapter = TypeAdapter(Filter)

class BaseFilters(_mappedFields):
    """Base filter validator."""

    model_config = ConfigDict(extra="allow")

    @model_validator(mode='after')
    def validate_extra_fields(self):
        """Validates extra fields (keys and values) to ensure they conform to safety and Filter type."""
        if not self.model_extra:
            return self

        for key, value in self.model_extra.items():
            # validate keys
            if not isinstance(key, str):
                 raise ValueError(f"Filter key must be a string, got {type(key).__name__} for key '{key}'.")

            if not ALLOWED_KEY_REGEX.fullmatch(key):
                 raise ValueError(
                     f"Filter key '{key}' contains invalid characters or format. "
                     f"Keys must match regex '{ALLOWED_KEY_REGEX.pattern}'."
                 )

            # validate values
            try:
                filter_adapter.validate_python(value)
            except (ValidationError, ValueError) as e:
                raise ValueError(
                    f"Filter key '{key}' has an invalid value. "
                    f"Expected a valid Filter type, but got {value}. "
                    f"Validation error: {e}"
                ) from e

        return self

    def std_conditions(self) -> Dict[str, List[Condition]]:
        """
        Maps filters to standard dict object format:
        - Key: Standard field names (from mapping or extra field keys)
        - Value: List[Condition] object(s)
        """
        # Helper function to convert a Filter value into a List[Condition]
        def _to_list_of_conditions(value: Filter) -> List[Condition]:
            if isinstance(value, list):
                # Check if it's already a list of Conditions
                if value and isinstance(value[0], Condition):
                    return value
                else:
                    # Assume it's a list of values for an IN operator
                    return [Condition(operator=Operator.IN, value=value)]
            else:
                # Assume it's a single value for an EQ operator
                return [Condition(operator=Operator.EQ, value=value)]

        standardized_filters: Dict[str, List[Condition]] = {}

        # Process defined fields that were set and are in the indexed keys
        defined_fields_to_process = self.model_fields_set & BaseNode.indexed_keys(self._flow)
        defined_filters = {
            field_name: _to_list_of_conditions(getattr(self, field_name))
            for field_name in defined_fields_to_process
        }
        standardized_filters.update(defined_filters)

        # Process extra fields
        if self.model_extra:
            extra_filters = {
                key: _to_list_of_conditions(value)
                for key, value in self.model_extra.items()
            }
            standardized_filters.update(extra_filters)

        return standardized_filters

    def std_dict(self) -> Dict[str, Any]:
        """
        Maps filters to standard dict format:
        - Key: Standard field names (from mapping or extra field keys)
        - Value: Primitive | List[Primitive]
        """
        conditions = self.std_conditions()
        conditions = {k: condition.value for k, v in conditions.items() for condition in v}

        return conditions

class ChatFilters(BaseFilters):
    """Metadata filter specific to `chat` flow."""
    _flow: Flow = PrivateAttr(Flow.CHAT)

    user_id: Optional[Filter] = Field(None, description="`Indexed` A user ID is an isolated search space.")
    chat_id: Optional[Filter] = Field(None, description="`Indexed` Metadata for filtering.")

class FileFilters(BaseFilters):
    """Metadata filter specific to `file` flow."""
    _flow: Flow = PrivateAttr(Flow.FILE)

    folder_id: Optional[Filter] = Field(None, description="`Indexed` A folder is an isolated search space. Not required only for read operations.")
    user_id: Optional[Filter] = Field(None, description="`Indexed` Metadata filter.")
    file_id: Optional[Filter] = Field(None, description="`Indexed` Metadata filter.")
    title: Optional[Filter] = Field(None, description="Metadata filter.")
    path: Optional[Filter] = Field(None, description="Metadata filter.")

Filters = ChatFilters | FileFilters

####################################
# --- Nodes and Relations ---
####################################

class NodeType(Enum):
    """Enum representing the possible types of nodes."""
    FACT = "fact"
    CHUNK = "chunk"
    TOPIC = "topic"
    ENTITY = "entity"
    def __str__(self):
        return self.value

class BaseNode(BaseModel):
    """Internal structure with all node properties."""
    node_type: NodeType
    flow: Flow
    content: str
    embedding: Optional[List[float]] = None
    created_at: Optional[int] = None
    updated_at: Optional[int] = None
    similarity: Optional[float] = None
    # indexed fields
    node_id: Optional[Id] = None
    user_id: Optional[Id] = None
    chat_id: Optional[Id] = None
    folder_id: Optional[Id] = None
    file_id: Optional[Id] = None
    # chunk node fields
    custom: Optional[Dict[str, Primitive]] = None

    # Non-db fields
    _descriptor: Optional[str] = PrivateAttr(None)
    _hash: str = PrivateAttr(None)

    max_name_chars: ClassVar[int] = 32

    @staticmethod
    def node_keys(node_type: NodeType) -> Set[str]:
        """Keys that represent data fields of the node, excluding private/class vars."""
        keys = set(BaseNode.model_fields.keys())
        if node_type != NodeType.CHUNK:
            keys -= {"custom"}

        return keys

    @staticmethod
    def indexed_keys(flow: Optional[Flow] = None) -> Set[str]:
        """Keys that are used for indexing."""
        k = {
            Flow.CHAT: {"user_id", "chat_id"},
            Flow.FILE: {"folder_id", "user_id", "file_id"},
        }

        if flow is None:
            return set.union(*k.values())
        else:
            return k[flow]

    @staticmethod
    def search_space(flow: Optional[Flow] = None) -> Set[str]:
        """Keys that are used for indexing."""
        k = {
            Flow.CHAT: {"user_id"},
            Flow.FILE: {"folder_id"},
        }

        if flow is None:
            return set.union(*k.values())
        else:
            return k[flow]

    def __init__(self, **data):
        super().__init__(**data)
        self._update_hash()

    def _update_hash(self) -> str:
        """Compute a hash based on the node's content and ids."""
        self._hash = hashlib.md5(f"{self.node_type}:{self.flow}:{self.content}".encode()).hexdigest()

    def __eq__(self, other):
        if not isinstance(other, BaseNode):
            return False
        return self._hash == other._hash

    @field_validator('content')
    @classmethod
    def strip_content(cls, v):
        if v is None:
            return None
        return v.strip()

    def __str__(self) -> str:
        return self.content or ""

    @field_validator('created_at', 'updated_at')
    @classmethod
    def to_millis(cls, value: Optional[int]) -> Optional[int]:
        """Validates if the timestamp is likely in seconds and converts it to milliseconds."""
        if value is None:
            return None

        if not isinstance(value, (int, float)):
             raise ValueError("Timestamp must be an integer or float")

        value = int(value)

        if value <= 1e8 or 1e16 <= value:
            raise ValueError("Invalid timestamp value")

        if 1e14 < value:
            return int(value/1000)
        elif 1e11 < value:
            return value
        elif 1e8 < value:
            return value * 1000

class ChatNode(BaseNode, _mappedFields):
    """Represents a basic unit of memory for a `chat` flow."""
    _flow: Flow = PrivateAttr(Flow.CHAT)

    user_id: Id = Field(description="`Indexed` A user ID is an isolated search space.")
    chat_id: Optional[Id] = Field(None, description="`Indexed` Metadata for filtering.")

class FileNode(BaseNode, _mappedFields):
    """Represents a basic unit of memory for a `file` flow."""
    _flow: Flow = PrivateAttr(Flow.FILE)

    path: Optional[str] = Field(None, description="Needed to process the file.")
    folder_id: Optional[Id] = Field(None, description="`Indexed` A folder ID is an isolated search space.")
    user_id: Optional[Id] = Field(None, description="`Indexed` Metadata for filtering.")
    file_id: Optional[Id] = Field(None, description="`Indexed` Metadata for filtering.")
    title: Optional[str] = Field(None, description="Metadata for filtering.")
    ichunk: Optional[int] = Field(None, description="Number of chunk in the file, base 0.")
    chunks: Optional[int] = Field(None, description="Total number of chunks in the file.")

Node = ChatNode | FileNode

class Relation(BaseModel):
    """Represents the basic memory unit."""
    fact: BaseNode = Field(description="The facts associated with the relation")
    chunk: BaseNode = Field(description="The chunk of the relation")
    topics: List[BaseNode] = Field(description="The topics of the relation")
    entities: List[BaseNode] = Field(description="The entities involved in the relation")

    def to_list(self) -> List[BaseNode]:
        """Convert the relation to a list of nodes.
        
        Returns a list in the order:
            [fact, chunk, topics..., entities...]
        """
        return [self.fact, self.chunk] + self.topics + self.entities
        
    def to_dicts(self, include: Set = None, exclude: Set = None, exclude_none=False) -> List[Dict[str, Any]]:
        """Convert the relation to a list of node dictionaries.
        
        Returns a list of dictionaries in the same order as to_list():
            [fact, chunk, topics..., entities...]
        """
        nodes = self.to_list()
        return [node.model_dump(mode='json', include=include, exclude=exclude, exclude_none=exclude_none) 
                for node in nodes]

    def to_id_nodes(self, include: Set = None, exclude: Set = None, exclude_none=False) -> Dict[str, Any]:
        """Convert the relation to a dictionary.
        
        Keys:
            n0: fact
            n1: chunk
            n*: topics and entities (n3, n4, ...)
        """
        node_dicts = self.to_dicts(include, exclude, exclude_none)
        
        return {f"n{i}": d for i, d in enumerate(node_dicts)}


####################################
# --- File operations ---
####################################

class FileStatus(Enum):
    """Enum for possible file processing states."""
    PENDING = 'pending'
    PROCESSING = 'processing'
    COMPLETED = 'completed'
    ERROR = 'error'
    CANCELLED = 'cancelled'
    def __str__(self):
        return self.value

class ChunkResult(BaseModel):
    """Represents the result of processing a single chunk."""
    chunk_index: int = Field(..., description="Index of the chunk within the file.")
    ids: Optional[List[str]] = Field(None, description="List of IDs generated for the upserted items in this chunk.")
    error: Optional[str] = Field(None, description="Error message if processing this chunk failed.")

class FileState(BaseModel):
    """Represents the processing state and cumulative result of a single file."""
    file_path: str = Field(..., description="Path to the file being processed.") # Added from FileUpsertResult
    task_id: Optional[str] = Field(None, description="ID of the task processing this file.")
    total_chunks: Optional[int] = Field(None, description="Total number of chunks expected for this file.")
    chunk_results: List[ChunkResult] = Field([], description="List of results for each processed chunk.") # Added from FileUpsertResult, now cumulative
    status: FileStatus = Field(FileStatus.PENDING, description="Current status of the file processing.")
    error: Optional[str] = Field(None, description="Error message if status is ERROR or CANCELLED.") # Kept from both

    # is_terminal is kept as a property
    @property
    def is_terminal(self) -> bool:
        """Returns True if the file is in a final state (completed, error, cancelled)."""
        return self.status in {FileStatus.COMPLETED, FileStatus.ERROR, FileStatus.CANCELLED}

    # Helper method to create an error state (similar to FileUpsertResult.error)
    @classmethod
    def create_error_state(cls, file_path: str, error_msg: str, task_id: Optional[str] = None) -> 'FileState':
        """Helper to create a FileState instance representing an error."""
        return cls(
            file_path=file_path,
            task_id=task_id,
            status=FileStatus.ERROR,
            error=error_msg,
            # total_chunks and chunk_results would likely be unknown or empty on a file-level error
            total_chunks=None,
            chunk_results=[]
        )

    # Helper method to update state with a new chunk result
    def update_with_chunk_result(self, chunk_result: ChunkResult) -> None:
        """Updates the state by adding a new chunk result and potentially updating status."""
        self.chunk_results.append(chunk_result)
        # Transition from PENDING to PROCESSING on the first chunk result
        if self.status == FileStatus.PENDING:
            self.status = FileStatus.PROCESSING
        # Transition to COMPLETED if total_chunks is known and all chunks are processed
        if self.total_chunks is not None and len(self.chunk_results) == self.total_chunks:
            # Note: Individual chunk errors don't change the file status from COMPLETED. File-level error is established by create_error_state()
            self.status = FileStatus.COMPLETED

class BatchStatusSummary(BaseModel):
    """Summary of the current batch processing status."""
    total_files: int
    files_done: int
    processing_files: int
    percentage: float
    elapsed_time: float
    remaining_time: Optional[float] = None
    file_states: Dict[str, FileState] = {}


####################################
# --- APIs and models ---
####################################

class UrlKeyModel(NamedTuple):
    url: str
    key: str
    model: str

class UrlKey(NamedTuple):
    url: str
    key: str


####################################
# --- Prompts ---
####################################

class Language(Enum):
    ENGLISH = "en"
    SPANISH = "es"
    FRENCH = "fr"
    GERMAN = "de"

    def __str__(self):
        return self.value

    @classmethod
    def _missing_(cls, value):
        """Handle variations like 'en-US'."""
        if isinstance(value, str):
            base_lang = value.split('-')[0].lower()
            for member in cls:
                if member.value == base_lang:
                    return member
        return super()._missing_(value)

class Translations(BaseModel):
    en: str
    es: str
    fr: str
    de: str

class Example(BaseModel):
    flow: Flow = Field(..., description="Type of example")
    example_input: Translations = Field(..., description="Example input translations")
    example_output: Translations = Field(..., description="Example output translations")

class Prompt(BaseModel):
    prompt_type: Literal["system", "user"] = Field(..., description="Type of prompt")
    prompt: Translations = Field(..., description="Prompt translations")
    examples: List[Example] = Field(..., description="Examples of inputs and outputs")

    def to_str(self, flow: Optional[Flow] = None, language: Language = Language.ENGLISH) -> str:
        r = getattr(self.prompt, language.value)
        examples = [
            f"[EXAMPLE_INPUT]\n{getattr(e.example_input, language.value)}\n[/EXAMPLE_INPUT]\n[EXAMPLE_OUTPUT]\n{getattr(e.example_output, language.value)}"
            for e in self.examples if not flow or e.flow == flow
        ]
        if not examples:
            return r

        r += "\n\n--- This is an example ---\n\n"
        r += "\n\n--- This is another example ---\n\n".join(examples)
        r += "\n\n--- This was an example ---\n\n"
        return r

####################################
# --- Helper functions ---
####################################

OUTPUT_CLASS_MAP: Dict[Type, Dict[Flow, Type]] = {
    BaseMetadata: {
        Flow.CHAT: ChatMetadata,
        Flow.FILE: FileMetadata,
    },
    BaseFilters: {
        Flow.CHAT: ChatFilters,
        Flow.FILE: FileFilters,
    },
    BaseNode: {
        Flow.CHAT: ChatNode,
        Flow.FILE: FileNode,
    },
}

def get_base_target(output_base: Type, flow: Optional[Flow]) -> Type:
    """Match the output class flow type to the input base class and flow."""
    flow_map = OUTPUT_CLASS_MAP.get(output_base)
    if flow_map and flow in flow_map:
        return flow_map[flow]

    return output_base




from ragchat.definitions import Prompt, Translations, Example

MSG_CLASSIFICATION = Prompt(prompt_type="system", prompt=Translations(

en="""
Task: Classify the INPUT as 'statement', 'question', 'none'.

Definitions:
- statement: A declaration, assertion or instruction.
- question: An interrogative message seeking information.
- none: The input is incoherent, empty or does not fit into any category.

Instructions:
- Don't answer questions or provide explanations, just classify the message.
- Write 'statement' or 'question' or 'none'
""",

es="""
Tarea: Clasifica la ENTRADA como 'statement', 'question' o 'none'.

Definiciones:
- statement: Una declaración, afirmación o instrucción.
- question: Un mensaje interrogativo que busca información.
- none: La entrada es incoherente, está vacía o no encaja en ninguna categoría.

Instrucciones:
- No respondas preguntas ni des explicaciones, solo clasifica el mensaje.
- Escribe 'statement' o 'question' o 'none'
""",

fr="""
Tâche : Classifier l'ENTRÉE comme 'statement', 'question' ou 'none'.

Définitions :
- statement : Une déclaration, une affirmation ou une instruction.
- question : Un message interrogatif cherchant à obtenir des informations.
- none : L'entrée est incohérente, vide ou ne correspond à aucune catégorie.

Instructions :
- Ne répondez pas aux questions et ne donnez pas d'explications, classez simplement le message.
- Écrivez 'statement', 'question' ou 'none'
""",

de="""
Aufgabe: Klassifiziere die EINGABE als 'statement', 'question' oder 'none'.

Definitionen:
- statement: Eine Aussage, Behauptung oder Anweisung.
- question: Eine Frage, die nach Informationen sucht.
- none: Die Eingabe ist unzusammenhängend, leer oder passt in keine Kategorie.

Anweisungen:
- Beantworte keine Fragen und gib keine Erklärungen, sondern klassifiziere nur die Nachricht.
- Schreibe 'statement', 'question' oder 'none'
""",

), examples=[

Example(flow="chat", example_input=Translations(
en="""Is it true that the Kingdom of Italy was proclaimed in 1861, while the French Third Republic was established in 1870 and therefore Italy is technically older?""",
es="""¿Es cierto que el Reino de Italia fue proclamado en 1861, mientras que la Tercera República Francesa se estableció en 1870 y por lo tanto Italia es técnicamente más antigua?""",
fr=""""Est-il vrai que le Royaume d'Italie a été proclamé en 1861, tandis que la Troisième République française a été établie en 1870 et que l'Italie est donc techniquement plus ancienne?""",
de="""Stimmt es, dass das Königreich Italien 1861 ausgerufen wurde, während die Dritte Französische Republik 1870 gegründet wurde und Italien somit technisch älter ist?""",
),
example_output=Translations(
en="question",
es="question",
fr="question",
de="question",
)),

Example(flow="chat", example_input=Translations(
en="""Paris is the capital of France. Rome is the capital of Italy.""",
es="""París es la capital de Francia. Roma es la capital de Italia.""",
fr=""""Paris est la capitale de la France. Rome est la capitale de l'Italie.""",
de="""Paris ist die Hauptstadt Frankreichs. Rom ist die Hauptstadt Italiens.""",
),
example_output=Translations(
en="statement",
es="statement",
fr="statement",
de="statement",
)),

Example(flow="chat", example_input=Translations(
en="""Florp wizzlebop cronkulated the splonk! Zizzleflap {}[]()<>:;"/|,.<>? drumblesquanch, but only if the quibberflitz jibberflops. Blorp???""",
es="""¡Florp wizzlebop cronkuleó el splonk! ¡Zizzleflap {}[]()<>:;"/|,.<>? drumblecuancha, pero solo si el quibberflitz jibberflops! ¡Blorp???""",
fr="""Florp wizzlebop a cronkulé le splonk ! Zizzleflap {}[]()<>:;"/|,.<>? drumblecuancha, mais seulement si le quibberflitz jibberflops. Blorp ???""",
de="""Florp wizzlebop hat den Splonk cronkuliert! Zizzleflap {}[]()<>:;"/|,.<>? drumblecuancha, aber nur wenn der quibberflitz jibberflops. Blorp???""",
),
example_output=Translations(
en="none",
es="none",
fr="none",
de="none",
)),
])

TOPICS = Prompt(prompt_type="system", prompt=Translations(

en="""
Task: Extract distinctive topics.

Instructions:
- Extract topics from the INPUT text
- The topics should be concise and serve as differentiators from similar texts
- If the INPUT is incoherent or invalid, list 'None'
- No explanations, just output in markdown format:
## Distinctive topics
- [topic]
""",

es="""
Tarea: Extraer temas distintivos.

Instrucciones:
- Extraer temas del texto INPUT
- Los temas deben ser concisos y servir como diferenciadores de textos similares
- Si el INPUT es incoherente o inválido, listar 'None'
- Nada de explicaciones, solo escribe en formato markdown:
## Temas distintivos
- [tema]
""",

fr="""
Tâche : Extraire des sujets distinctifs.

Instructions :
- Extraire les sujets du texte INPUT
- Les sujets doivent être concis et servir de différenciateurs par rapport à des textes similaires
- Si l'INPUT est incohérent ou invalide, lister 'None'
- Pas d'explications, juste la sortie au format markdown :
## Sujets distinctifs
- [sujet]
""",

de="""
Aufgabe: Unterscheidende Themen extrahieren.

Anweisungen:
- Themen aus dem INPUT Text extrahieren
- Die Themen sollten prägnant sein und als Unterscheidungsmerkmale zu ähnlichen Texten dienen
- Wenn der INPUT inkohärent oder ungültig ist, 'None' auflisten
- Keine Erklärungen, nur die Ausgabe im Markdown-Format:
## Unterscheidende Themen
- [Thema]
"""
), examples=[

Example(flow="chat", example_input=Translations(
en="""
I had dinner with my family last night. My husband cooked lasagna, which is my daughter's favorite dish. Our son brought his girlfriend, who we met for the first time. She brought a lovely bottle of wine that paired perfectly with the meal.

Optional topics:
- Family dinner
- Breakfast with friends
""",
es="""
Cené con mi familia anoche. Mi esposo cocinó lasaña, que es el plato favorito de mi hija. Nuestro hijo trajo a su novia, a quien conocimos por primera vez. Ella trajo una botella de vino estupenda que maridó perfectamente con la comida.

Temas opcionales:
- Cena familiar
- Desayuno con amigos
""",
fr="""
J'ai dîné avec ma famille hier soir. Mon mari a cuisiné des lasagnes, qui est le plat préféré de ma fille. Notre fils a amené sa petite amie, que nous rencontrions pour la première fois. Elle a apporté une jolie bouteille de vin qui s'est parfaitement accordée avec le repas.

Sujets optionnels:
- Dîner en famille
- Petit-déjeuner entre amis
""",
de="""
Ich habe gestern Abend mit meiner Familie zu Abend gegessen. Mein Mann hat Lasagne gekocht, was das Lieblingsgericht meiner Tochter ist. Unser Sohn hat seine Freundin mitgebracht, die wir zum ersten Mal kennenlernten. Sie hat eine schöne Flasche Wein mitgebracht, die perfekt zum Essen passte.

Optionale Themen:
- Familienessen
- Frühstück mit Freunden
""",
), example_output=Translations(
en="""
## Distinctive topics
- Family dinner
- Meeting son's girlfriend (first time)
""",
es="""
## Temas distintivos
- Cena familiar
- Conocer a la novia del hijo (primera vez)
""",
fr="""
## Sujets distinctifs
- Dîner en famille
- Rencontrer la petite amie du fils (première fois)
""",
de="""
## Unterscheidende Themen
- Familienessen
- Die Freundin des Sohnes kennenlernen (erstes Mal)
""",
)),

Example(flow="file", example_input=Translations(
en="""
Eukaryotic cells, the building blocks of complex life, are thought to have evolved from simpler prokaryotic cells through a process called endosymbiosis, where one prokaryotic cell engulfed another, leading to the development of membrane-bound organelles.
Eukaryotic cells have internal organization with a plasma membrane, cytoplasm containing organelles like mitochondria (energy), endoplasmic reticulum (synthesis), and a nucleus (genetic information) enclosed by a nuclear envelope.

Optional topics:
- Prokaryotic Cells
- Evolution of Eukaryotic Cells (early evolution from Prokaryotes)
""",
es="""
Se cree que las células eucariotas, los bloques de construcción de la vida compleja, evolucionaron a partir de células procariotas más simples a través de un proceso llamado endosimbiosis, donde una célula procariota engulló a otra, lo que llevó al desarrollo de orgánulos unidos a la membrana.
Las células eucariotas tienen organización interna con una membrana plasmática, citoplasma que contiene orgánulos como las mitocondrias (energía), retículo endoplasmático (síntesis) y un núcleo (información genética) encerrado por una envoltura nuclear.

Temas opcionales:
- Células procariotas
- Evolución de las células eucariotas (evolución temprana a partir de procariotas)
""",
fr="""
Les cellules eucaryotes, les éléments constitutifs de la vie complexe, sont considérées comme ayant évolué à partir de cellules procaryotes plus simples par un processus appelé endosymbiose, où une cellule procaryote en a englouti une autre, conduisant au développement d'organites liés à la membrane.
Les cellules eucaryotes ont une organisation interne avec une membrane plasmique, un cytoplasme contenant des organites comme les mitochondries (énergie), le réticulum endoplasmique (synthèse), et un noyau (information génétique) entouré d'une enveloppe nucléaire.

Sujets optionnels:
- Cellules procaryotes
- Évolution des cellules eucaryotes (évolution précoce à partir des procaryotes)
""",
de="""
Eukaryotische Zellen, die Bausteine komplexen Lebens, sollen sich aus einfacheren prokaryotischen Zellen durch einen Prozess namens Endosymbiose entwickelt haben, bei dem eine prokaryotische Zelle eine andere verschlang, was zur Entwicklung membranumhüllter Organellen führte.
Eukaryotische Zellen haben eine innere Organisation mit einer Plasmamembran, Zytoplasma, das Organellen wie Mitochondrien (Energie), endoplasmatisches Retikulum (Synthese) und einen Zellkern (genetische Information), umschlossen von einer Kernhülle, enthält.

Optionale Themen:
- Prokaryotische Zellen
- Evolution der Eukaryotischen Zellen (frühe Evolution aus Prokaryoten)
""",
), example_output=Translations(
en="""
## Distinctive topics
- Evolution of Eukaryotic Cells (early evolution from Prokaryotes)
- Eukaryotic Cells
""",
es="""
## Temas distintivos
- Evolución de las células eucariotas (evolución temprana a partir de procariotas)
- Células eucariotas
""",
fr="""
## Sujets distinctifs
- Évolution des cellules eucaryotes (évolution précoce à partir des procaryotes)
- Cellules eucaryotes
""",
de="""
## Unterscheidende Themen
- Evolution der Eukaryotischen Zellen (frühe Evolution aus Prokaryoten)
- Eukaryotische Zellen
""",
)),
])

TOPIC_FACTS = Prompt(prompt_type="system", prompt=Translations(

en="""
Task: Extract all topics and facts.

Instructions:
- Strictly use the topics from INPUT as headings, do not modify them, do not add new ones
- Extract all facts from the INPUT, facts should be concise and unambiguous
- List the facts under their respective topic, facts may repeat across topics
- Use names instead of pronouns
- If the INPUT is incoherent or invalid, list 'None'
- No explanations, just output in markdown format:
## [topic]
- [fact]
""",

es="""
Tarea: Extraer todos los temas y hechos.

Instrucciones:
- Utiliza estrictamente los temas del INPUT como encabezados, no los modifiques, no añadas nuevos
- Extrae todos los hechos del INPUT, los hechos deben ser concisos y no ambiguos
- Lista los hechos bajo su tema respectivo, los hechos pueden repetirse entre temas
- Usa nombres en lugar de pronombres
- Si el INPUT es incoherente o inválido, lista 'None'
- Nada de explicaciones, solo escribe en formato markdown:
## [tema]
- [hecho]
""",

fr="""
Tâche: Extraire tous les sujets et faits.

Instructions:
- Utilise strictement les sujets de l'INPUT comme titres, ne les modifie pas, n'en ajoute pas de nouveaux
- Extrais tous les faits de l'INPUT, les faits doivent être concis et non ambigus
- Liste les faits sous leur sujet respectif, les faits peuvent se répéter entre les sujets
- Utilise des noms au lieu de pronoms
- Si l'INPUT est incohérent ou invalide, liste 'None'
- Pas d'explications, juste la sortie au format markdown :
## [sujet]
- [fait]
""",

de="""
Aufgabe: Extrahieren Sie alle Themen und Fakten.

Anweisungen:
- Verwenden Sie strikt die Themen aus dem INPUT als Überschriften, ändern Sie sie nicht, fügen Sie keine neuen hinzu
- Extrahieren Sie alle Fakten aus dem INPUT, Fakten sollten prägnant und eindeutig sein
- Listen Sie die Fakten unter ihrem jeweiligen Thema auf, Fakten können sich über Themen hinweg wiederholen
- Verwenden Sie Namen anstelle von Pronomen
- Wenn der INPUT inkohärent oder ungültig ist, listen Sie 'None' auf
- Keine Erklärungen, geben Sie nur die Ausgabe im Markdown-Format aus:
## [Thema]
- [Fakt]
""",

), examples=[

Example(flow="chat", example_input=Translations(
en="""
current timestamp: 2023-02-12 13:30

# Topics
## Family dinner
## Meeting son's girlfriend (first time)

I had dinner with my family last night. My husband cooked lasagna, which is my daughter's favorite dish. Our son brought his girlfriend, who we were meeting for the first time. She brought a lovely bottle of wine that paired perfectly with the meal.
""",
es="""
current timestamp: 2023-02-12 13:30

# Temas
## Cena familiar
## Conocer a la novia del hijo (primera vez)

Anoche cené con mi familia. Mi esposo cocinó lasaña, que es el plato favorito de mi hija. Nuestro hijo trajo a su novia, a quien conocíamos por primera vez. Ella trajo una botella de vino encantadora que maridó perfectamente con la comida.
""",
fr="""
current timestamp: 2023-02-12 13:30

# Sujets
## Dîner en famille
## Rencontrer la petite amie du fils (première fois)

J'ai dîné avec ma famille hier soir. Mon mari a cuisiné des lasagnes, qui est le plat préféré de ma fille. Notre fils a amené sa petite amie, que nous rencontrions pour la première fois. Elle a apporté une charmante bouteille de vin qui s'est parfaitement accordée avec le repas.
""",
de="""
current timestamp: 2023-02-12 13:30

# Themen
## Familienessen
## Die Freundin des Sohnes kennenlernen (erstes Mal)

Gestern Abend habe ich mit meiner Familie zu Abend gegessen. Mein Mann hat Lasagne gekocht, das Lieblingsgericht meiner Tochter. Unser Sohn hat seine Freundin mitgebracht, die wir zum ersten Mal trafen. Sie hat eine schöne Flasche Wein mitgebracht, die perfekt zum Essen passte.
""",
), example_output=Translations(
en="""
## Family dinner
- The user had dinner with the user's family (last night from 2023-02-12 13:30).
- The user's husband cooked lasagna.
- Lasagna is the user's daughter's favorite dish.
- The user's son brought his girlfriend to the family dinner.
- The girlfriend brought a lovely bottle of wine.
- The wine paired perfectly with the meal.
## Meeting son's girlfriend (first time)
- The user met the user's son's girlfriend (first time).
- The user's son brought his girlfriend to the family dinner.
- The girlfriend brought a lovely bottle of wine.
""",
es="""
## Cena familiar
- El usuario/a cenó con su familia (anoche, desde el 12-02-2023 13:30).
- El marido del usuario/a cocinó lasaña.
- La lasaña es el plato favorito de la hija del usuario/a.
- El hijo del usuario/a trajo a su novia a la cena familiar.
- La novia trajo una botella de vino encantadora.
- El vino maridó perfectamente con la comida.
## Conocer a la novia del hijo (primera vez)
- El usuario/a conoció a la novia de su hijo (primera vez).
- El hijo del usuario/a trajo a su novia a la cena familiar.
- La novia trajo una botella de vino encantadora.
""",
fr="""
## Dîner de famille
- L'utilisateur/trice a dîné avec sa famille (hier soir à partir du 12-02-2023 13:30).
- Le mari de l'utilisateur/trice a cuisiné des lasagnes.
- Les lasagnes sont le plat préféré de la fille de l'utilisateur/trice.
- Le fils de l'utilisateur/trice a amené sa petite amie au dîner de famille.
- La petite amie a apporté une charmante bouteille de vin.
- Le vin s'est parfaitement accordé avec le repas.
## Rencontre avec la petite amie du fils (première fois)
- L'utilisateur/trice a rencontré la petite amie de son fils (première fois).
- Le fils de l'utilisateur/trice a amené sa petite amie au dîner de famille.
- La petite amie a apporté une charmante bouteille de vin.
""",
de="""
## Familienessen
- Der Benutzer/in hat mit seiner Familie zu Abend gegessen (gestern Abend ab 13:30 Uhr am 12.02.2023).
- Der Ehemann des Benutzer/in hat Lasagne gekocht.
- Lasagne ist das Lieblingsgericht der Tochter des Benutzer/in.
- Der Sohn des Benutzer/in hat seine Freundin zum Familienessen mitgebracht.
- Die Freundin hat eine schöne Flasche Wein mitgebracht.
- Der Wein passte perfekt zum Essen.
## Treffen mit der Freundin des Sohnes (erstes Mal)
- Der Benutzer/in hat die Freundin seines/ihres Sohnes getroffen (erstes Mal).
- Der Sohn des Benutzer/in hat seine Freundin zum Familienessen mitgebracht.
- Die Freundin hat eine schöne Flasche Wein mitgebracht.
""",
)),

Example(flow="file", example_input=Translations(
en="""
source: "Eukaryotes, Origin of". Encyclopedia of Biodiversity. Vol. 2.

# Topics
## Evolution of Eukaryotic Cells (early evolution from Prokaryotes)
## Eukaryotic Cells

Eukaryotic cells, the building blocks of complex life, are thought to have evolved from simpler prokaryotic cells through a process called endosymbiosis, where one prokaryotic cell engulfed another, leading to the development of membrane-bound organelles.
Eukaryotic cells have internal organization with a plasma membrane, cytoplasm containing organelles like mitochondria (energy), endoplasmic reticulum (synthesis), and a nucleus (genetic information) enclosed by a nuclear envelope.
""",
es="""
fuente: "Eucariotas, Origen de". Enciclopedia de la Biodiversidad. Vol. 2.

# Temas
## Evolución de las Células Eucariotas (evolución temprana a partir de Procariotas)
## Células Eucariotas

Las células eucariotas, los componentes básicos de la vida compleja, se cree que evolucionaron a partir de células procariotas más simples a través de un proceso llamado endosimbiosis, donde una célula procariota engulló a otra, lo que llevó al desarrollo de orgánulos unidos a membranas.
Las células eucariotas tienen organización interna con una membrana plasmática, citoplasma que contiene orgánulos como mitocondrias (energía), retículo endoplasmático (síntesis) y un núcleo (información genética) encerrado por una envoltura nuclear.
""",
fr="""
source : "Eucaryotes, Origine des". Encyclopédie de la Biodiversité. Vol. 2.

# Sujets
## Évolution des Cellules Eucaryotes (évolution précoce à partir des Procaryotes)
## Cellules Eucaryotes

Les cellules eucaryotes, les éléments constitutifs de la vie complexe, sont censées avoir évolué à partir de cellules procaryotes plus simples par un processus appelé endosymbiose, où une cellule procaryote en a englouti une autre, conduisant au développement d'organites liés à une membrane.
Les cellules eucaryotes ont une organisation interne avec une membrane plasmique, un cytoplasme contenant des organites comme les mitochondries (énergie), le réticulum endoplasmique (synthèse) et un noyau (information génétique) entouré d'une enveloppe nucléaire.
""",
de="""
Quelle: "Eukaryoten, Ursprung der". Enzyklopädie der Biodiversität. Bd. 2.

# Themen
## Evolution der Eukaryotischen Zellen (frühe Evolution aus Prokaryoten)
## Eukaryotische Zellen

Eukaryotische Zellen, die Bausteine komplexen Lebens, sollen sich aus einfacheren prokaryotischen Zellen durch einen Prozess namens Endosymbiose entwickelt haben, bei dem eine prokaryotische Zelle eine andere verschlang, was zur Entwicklung membrangebundener Organellen führte.
Eukaryotische Zellen haben eine innere Organisation mit einer Plasmamembran, Zytoplasma, das Organellen wie Mitochondrien (Energie), endoplasmatisches Retikulum (Synthese) und einen von einer Kernhülle umschlossenen Zellkern (genetische Information) enthält.
""",
), example_output=Translations(
en="""
## Evolution of Eukaryotic Cells (early evolution from Prokaryotes)
- Eukaryotic cells are thought to have evolved from simpler prokaryotic cells through a process called endosymbiosis.
- In endosymbiosis, one prokaryotic cell engulfed another.
- Endosymbiosis led to the development of membrane-bound organelles.
## Eukaryotic Cells
- Eukaryotic cells are the building blocks of complex life.
- Eukaryotic cells have internal organization with a plasma membrane and cytoplasm.
- Cytoplasm contains organelles like mitochondria (energy), endoplasmic reticulum (synthesis) and nucleus (genetic information).
- The nucleus is enclosed by a nuclear envelope.
""",
es="""
## Evolución de las Células Eucariotas (evolución temprana a partir de Procariotas)
- Se cree que las células eucariotas evolucionaron a partir de células procariotas más simples a través de un proceso llamado endosimbiosis.
- En la endosimbiosis, una célula procariota engulló a otra.
- La endosimbiosis condujo al desarrollo de orgánulos unidos a membranas.
## Células Eucariotas
- Las células eucariotas son los componentes básicos de la vida compleja.
- Las células eucariotas tienen organización interna con una membrana plasmática y citoplasma.
- El citoplasma contiene orgánulos como mitocondrias (energía), retículo endoplasmático (síntesis) y núcleo (información genética).
- El núcleo está rodeado por una envoltura nuclear.
""",
fr="""
## Évolution des Cellules Eucaryotes (évolution précoce à partir des Procaryotes)
- On pense que les cellules eucaryotes ont évolué à partir de cellules procaryotes plus simples par un processus appelé endosymbiose.
- Dans l'endosymbiose, une cellule procaryote en a englouti une autre.
- L'endosymbiose a conduit au développement d'organites liés à des membranes.
## Cellules Eucaryotes
- Les cellules eucaryotes sont les éléments constitutifs de la vie complexe.
- Les cellules eucaryotes ont une organisation interne avec une membrane plasmique et un cytoplasme.
- Le cytoplasme contient des organites comme les mitochondries (énergie), le réticulum endoplasmique (synthèse) et le noyau (information génétique).
- Le noyau est entouré d'une enveloppe nucléaire.
""",
de="""
## Evolution der Eukaryotischen Zellen (frühe Evolution aus Prokaryoten)
- Es wird angenommen, dass sich eukaryotische Zellen aus einfacheren prokaryotischen Zellen durch einen Prozess namens Endosymbiose entwickelt haben.
- Bei der Endosymbiose verschlang eine prokaryotische Zelle eine andere.
- Die Endosymbiose führte zur Entwicklung von membranumschlossenen Organellen.
## Eukaryotische Zellen
- Eukaryotische Zellen sind die Bausteine komplexen Lebens.
- Eukaryotische Zellen haben eine innere Organisation mit einer Plasmamembran und Zytoplasma.
- Das Zytoplasma enthält Organellen wie Mitochondrien (Energie), Endoplasmatisches Retikulum (Synthese) und den Zellkern (genetische Information).
- Der Zellkern ist von einer Kernhülle umschlossen.
""",
)),
])

FACT_ENTITIES = Prompt(prompt_type="system", prompt=Translations(

en="""
Task: Extract facts and their corresponding entities.

Instructions:
- Strictly use the INPUT facts as headings, do not change them, do not add new ones
- For each fact, extract every entity (i.e. subjects, objects, events, concepts, etc.)
- Entities must be written with format: `name (type)`
- Entities may repeat across facts
- Use names instead of pronouns
- No explanations, just output in markdown format:
## [fact]
- name (type)
- name (type)
""",

es="""
Tarea: Extraer hechos y sus entidades correspondientes.

Instrucciones:
- Usar estrictamente los hechos del INPUT como encabezados, no cambiarlos, no añadir nuevos
- Para cada hecho, extraer cada entidad (es decir, sujetos, objetos, eventos, conceptos, etc.)
- Las entidades deben escribirse con el formato: `nombre (tipo)`
- Las entidades pueden repetirse entre hechos
- Usar nombres en lugar de pronombres
- Nada de explicaciones, solo escribe en formato markdown:
## [hecho]
- nombre (tipo)
- nombre (tipo)
""",

fr="""
Tâche: Extraire les faits et leurs entités correspondantes.

Instructions:
- Utiliser strictement les faits de l'INPUT comme titres, ne pas les modifier, ne pas en ajouter de nouveaux
- Pour chaque fait, extraire chaque entité (c'est-à-dire sujets, objets, événements, concepts, etc.)
- Les entités doivent être écrites au format: `nom (type)`
- Les entités peuvent se répéter entre les faits
- Utiliser des noms au lieu de pronoms
- Pas d'explications, juste la sortie au format markdown :
## [fait]
- nom (type)
- nom (type)
""",

de="""
Aufgabe: Fakten und ihre entsprechenden Entitäten extrahieren.

Anweisungen:
- Verwenden Sie strikt die INPUT-Fakten als Überschriften, ändern Sie sie nicht, fügen Sie keine neuen hinzu
- Extrahieren Sie für jeden Fakt jede Entität (d.h. Subjekte, Objekte, Ereignisse, Konzepte usw.)
- Entitäten müssen im Format geschrieben werden: `Name (Typ)`
- Entitäten können sich über Fakten hinweg wiederholen
- Verwenden Sie Namen anstelle von Pronomen
- Keine Erklärungen, nur die Ausgabe im Markdown-Format:
## [Fakt]
- Name (Typ)
- Name (Typ)
""",

), examples=[

Example(flow="chat", example_input=Translations(
en="""
current timestamp: 2023-02-12 13:30

# Facts
## User had dinner with the family last night.
## The husband of user cooked lasagna.
## Lasagna is the favorite dish of the daughter of user.
## The son of user brought his girlfriend.
## The family of user was meeting the girlfriend of the son of user for the first time.

Optional entities:
- User (current user)
- Husband (spouse of user)
- Bob (neighbor of user)
""",
es="""
tiempo actual: 2023-02-12 13:30

# Hechos
## El usuario/a cenó con la familia anoche.
## El marido del usuario/a cocinó lasaña.
## La lasaña es el plato favorito de la hija del usuario/a.
## El hijo del usuario/a trajo a su novia.
## La familia del usuario/a conoció a la novia del hijo del usuario/a por primera vez.

Entidades opcionales:
- Usuario/a (usuario/a actual)
- Marido (cónyuge del usuario/a)
- Bob (vecino del usuario/a)
""",
fr="""
horodatage actuel : 2023-02-12 13:30

# Faits
## L'utilisateur/trice a dîné avec la famille hier soir.
## Le mari de l'utilisateur/trice a cuisiné des lasagnes.
## Les lasagnes sont le plat préféré de la fille de l'utilisateur/trice.
## Le fils de l'utilisateur/trice a amené sa petite amie.
## La famille de l'utilisateur/trice rencontrait la petite amie du fils de l'utilisateur/trice pour la première fois.

Entités optionnelles :
- Utilisateur/trice (utilisateur/trice actuel/le)
- Mari (conjoint de l'utilisateur/trice)
- Bob (voisin de l'utilisateur/trice)
""",
de="""
aktueller Zeitstempel: 2023-02-12 13:30

# Fakten
## Der Benutzer/in hat gestern Abend mit der Familie zu Abend gegessen.
## Der Ehemann des Benutzers/in hat Lasagne gekocht.
## Lasagne ist das Lieblingsgericht der Tochter des Benutzers.
## Der Sohn des Benutzers/in hat seine Freundin mitgebracht.
## Die Familie des Benutzers/in hat die Freundin des Sohnes des Benutzers/in zum ersten Mal getroffen.

Optionale Entitäten:
- Benutzer/in (aktueller Benutzer/in)
- Ehemann (Ehepartner des Benutzers)
- Bob (Nachbar des Benutzers)
""",
), example_output=Translations(
en="""
## User had dinner with the family last night.
- User (current user)
- Family (relatives of user)
- Dinner (event last night from 2023-02-12 13:30)

## The husband of user cooked lasagna.
- Husband (spouse of user)
- Lasagna (Italian pasta dish)

## Lasagna is the favorite dish of the daughter of user.
- Lasagna (Italian pasta dish)
- Daughter (female child of user)

## The son of user brought his girlfriend.
- Son (male child of user)
- Girlfriend (romantic partner of son of user)

## The family of user was meeting the girlfriend of the son of user for the first time.
- Family (relatives of user)
- Girlfriend (romantic partner of son of user)
""",
es="""
## El usuario/a cenó con la familia anoche.
- Usuario/a (usuario/a actual)
- Familia (parientes del usuario/a)
- Cena (evento de anoche del 12-02-2023 13:30)

## El marido del usuario/a cocinó lasaña.
- Marido (cónyuge del usuario/a)
- Lasaña (plato de pasta italiano)

## La lasaña es el plato favorito de la hija del usuario/a.
- Lasaña (plato de pasta italiano)
- Hija (hija del usuario/a)

## El hijo del usuario/a trajo a su novia.
- Hijo (hijo del usuario/a)
- Novia (pareja sentimental del hijo del usuario/a)

## La familia del usuario/a conoció a la novia del hijo del usuario/a por primera vez.
- Familia (parientes del usuario/a)
- Novia (pareja sentimental del hijo del usuario/a)
""",
fr="""
## L'utilisateur/trice a dîné avec la famille hier soir.
- Utilisateur/trice (utilisateur/trice actuel/le)
- Famille (proches de l'utilisateur/trice)
- Dîner (événement d'hier soir du 12-02-2023 13:30)

## Le mari de l'utilisateur/trice a cuisiné des lasagnes.
- Mari (conjoint de l'utilisateur/trice)
- Lasagnes (plat de pâtes italien)

## Les lasagnes sont le plat préféré de la fille de l'utilisateur/trice.
- Lasagnes (plat de pâtes italien)
- Fille (enfant de sexe féminin de l'utilisateur/trice)

## Le fils de l'utilisateur/trice a amené sa petite amie.
- Fils (enfant de sexe masculin de l'utilisateur/trice)
- Petite amie (partenaire romantique du fils de l'utilisateur/trice)

## La famille de l'utilisateur/trice rencontrait la petite amie du fils de l'utilisateur/trice pour la première fois.
- Famille (proches de l'utilisateur/trice)
- Petite amie (partenaire romantique du fils de l'utilisateur/trice)
""",
de="""
## Der Benutzer/in hat gestern Abend mit der Familie zu Abend gegessen.
- Benutzer/in (aktueller Benutzer/in)
- Familie (Verwandte des/der Nutzers/in)
- Abendessen (Ereignis von gestern Abend vom 12.02.2023 13:30)

## Der Ehemann des Benutzers/in hat Lasagne gekocht.
- Ehemann (Ehepartner des/der Nutzers/in)
- Lasagne (italienisches Nudelgericht)

## Lasagne ist das Lieblingsgericht der Tochter des Benutzers.
- Lasagne (italienisches Nudelgericht)
- Tochter (weibliches Kind des/der Nutzers/in)

## Der Sohn des Benutzers/in hat seine Freundin mitgebracht.
- Sohn (männliches Kind des/der Nutzers/in)
- Freundin (romantische Partnerin des Sohnes des/der Nutzers/in)

## Die Familie des Benutzers/in hat die Freundin des Sohnes des Benutzers/in zum ersten Mal getroffen.
- Familie (Verwandte des/der Nutzers/in)
- Freundin (romantische Partnerin des Sohnes des/der Nutzers/in)
""",
)),

Example(flow="file", example_input=Translations(
en="""
source: "Eukaryotes, Origin of". Encyclopedia of Biodiversity. Vol. 2.

# Facts
## Eukaryotic cells are the building blocks of complex life.
## Eukaryotic cells have internal organization with a plasma membrane and cytoplasm.
## Cytoplasm contains organelles like mitochondria (energy), endoplasmic reticulum (synthesis) and nucleus (genetic information).
## The nucleus is enclosed by a nuclear envelope.

Optional entities:
- Prokariote (cell without membrane-bound nucleus)
- Plasma membrane (outer layer of cell)
- Cytoplasm (internal cell fluid)
- Genetic material (DNA)
""",
es="""
fuente: "Eucariotas, Origen de". Enciclopedia de la Biodiversidad. Vol. 2.

# Hechos
## Las células eucariotas son los componentes básicos de la vida compleja.
## Las células eucariotas tienen organización interna con una membrana plasmática y citoplasma.
## El citoplasma contiene orgánulos como mitocondrias (energía), retículo endoplasmático (síntesis) y núcleo (información genética).
## El núcleo está rodeado por una envoltura nuclear.

Entidades opcionales:
- Procariota (célula sin núcleo delimitado por membrana)
- Membrana plasmática (capa externa de la célula)
- Citoplasma (fluido interno de la célula)
- Material genético (ADN)
""",
fr="""
source : "Eucaryotes, Origine des". Encyclopédie de la Biodiversité. Vol. 2.

# Faits
## Les cellules eucaryotes sont les éléments constitutifs de la vie complexe.
## Les cellules eucaryotes ont une organisation interne avec une membrane plasmique et un cytoplasme.
## Le cytoplasme contient des organites comme les mitochondries (énergie), le réticulum endoplasmique (synthèse) et le noyau (information génétique).
## Le noyau est entouré d'une enveloppe nucléaire.

Entités optionnelles :
- Procaryote (cellule sans noyau délimité par une membrane)
- Membrane plasmique (couche externe de la cellule)
- Cytoplasme (fluide interne de la cellule)
- Matériel génétique (ADN)
""",
de="""
Quelle: "Eukaryoten, Ursprung der". Enzyklopädie der Biodiversität. Bd. 2.

# Fakten
## Eukaryotische Zellen sind die Bausteine komplexen Lebens.
## Eukaryotische Zellen haben eine innere Organisation mit einer Plasmamembran und Zytoplasma.
## Das Zytoplasma enthält Organellen wie Mitochondrien (Energie), endoplasmatisches Retikulum (Synthese) und Zellkern (genetische Information).
## Der Zellkern ist von einer Kernhülle umschlossen.

Optionale Entitäten:
- Prokaryot (Zelle ohne membranumschlossenen Zellkern)
- Plasmamembran (äußere Zellschicht)
- Zytoplasma (innere Zellflüssigkeit)
- Genetisches Material (DNA)
""",
), example_output=Translations(
en="""
## Eukaryotic cells are the building blocks of complex life.
- Eukaryotic cells (cell type)
- complex life (concept)

## Eukaryotic cells have internal organization with a plasma membrane and cytoplasm.
- Eukaryotic cells (cell type)
- internal organization (cellular feature)
- Plasma membrane (outer layer of cell)
- Cytoplasm (internal cell fluid)

## Cytoplasm contains organelles like mitochondria (energy), endoplasmic reticulum (synthesis) and nucleus (genetic information).
- Cytoplasm (internal cell fluid)
- organelles (cellular component)
- Mitochondria (energy-producing organelle)
- Endoplasmic reticulum (synthesis organelle)
- Nucleus (genetic information center)

## The nucleus is enclosed by a nuclear envelope.
- Nucleus (genetic information center)
- nuclear envelope (nucleus enclosure)
""",
es="""
## Las células eucariotas son los componentes básicos de la vida compleja.
- Células eucariotas (tipo de célula)
- vida compleja (concepto)

## Las células eucariotas tienen organización interna con una membrana plasmática y citoplasma.
- Células eucariotas (tipo de célula)
- organización interna (característica celular)
- Membrana plasmática (capa externa de la célula)
- Citoplasma (fluido celular interno)

## El citoplasma contiene orgánulos como mitocondrias (energía), retículo endoplasmático (síntesis) y núcleo (información genética).
- Citoplasma (fluido celular interno)
- orgánulos (componente celular)
- Mitocondrias (orgánulo productor de energía)
- Retículo endoplasmático (orgánulo de síntesis)
- Núcleo (centro de información genética)

## El núcleo está rodeado por una envoltura nuclear.
- Núcleo (centro de información genética)
- envoltura nuclear (recinto del núcleo)
""",
fr="""
## Les cellules eucaryotes sont les éléments constitutifs de la vie complexe.
- Cellules eucaryotes (type de cellule)
- vie complexe (concept)

## Les cellules eucaryotes ont une organisation interne avec une membrane plasmique et un cytoplasme.
- Cellules eucaryotes (type de cellule)
- organisation interne (caractéristique cellulaire)
- Membrane plasmique (couche externe de la cellule)
- Cytoplasme (fluide cellulaire interne)

## Le cytoplasme contient des organites comme les mitochondries (énergie), le réticulum endoplasmique (synthèse) et le noyau (information génétique).
- Cytoplasme (fluide cellulaire interne)
- organites (composant cellulaire)
- Mitochondries (organite producteur d'énergie)
- Réticulum endoplasmique (organite de synthèse)
- Noyau (centre d'information génétique)

## Le noyau est entouré d'une enveloppe nucléaire.
- Noyau (centre d'information génétique)
- enveloppe nucléaire (enceinte du noyau)
""",
de="""
## Eukaryotische Zellen sind die Bausteine komplexen Lebens.
- Eukaryotische Zellen (Zelltyp)
- komplexes Leben (Konzept)

## Eukaryotische Zellen haben eine innere Organisation mit einer Plasmamembran und Zytoplasma.
- Eukaryotische Zellen (Zelltyp)
- innere Organisation (zelluläres Merkmal)
- Plasmamembran (äußere Zellschicht)
- Zytoplasma (innere Zellflüssigkeit)

## Das Zytoplasma enthält Organellen wie Mitochondrien (Energie), endoplasmatisches Retikulum (Synthese) und Zellkern (genetische Information).
- Zytoplasma (innere Zellflüssigkeit)
- Organellen (zellulärer Bestandteil)
- Mitochondrien (energieproduzierende Organelle)
- Endoplasmatisches Retikulum (Synthese-Organelle)
- Zellkern (Zentrum genetischer Information)

## Der Zellkern ist von einer Kernhülle umschlossen.
- Zellkern (Zentrum genetischer Information)
- Kernhülle (Umschließung des Zellkerns)
""",
)),
])

ENTITY_DEFINITIONS = Prompt(prompt_type="system", prompt=Translations(

en="""
Task: Specify what the indefinite headings refer to.

Instructions:
- Select from the allowed headings only the ones that are indefinite (i.e. unclear, general, ambiguous)
- Select means select, do not modify the selected headings, do not add new ones
- For each selected heading, list one by one every entity the heading refers to
- Entities must be written with format: `name (type)`
- Entity format is one entity with a single set of parentheses
- Entities may repeat across headings
- If not applicable, list 'None'
- No explanations, just output in markdown format:
## [heading]
- name (type)
""",

es="""
Tarea: Especificar a qué se refieren los encabezados indefinidos.

Instrucciones:
- Seleccionar de los encabezados permitidos solo aquellos que sean indefinidos (es decir, poco claros, generales, ambiguos)
- Seleccionar significa seleccionar, no modificar los encabezados seleccionados, no añadir nuevos
- Para cada encabezado seleccionado, listar una por una cada entidad a la que se refiere el encabezado
- Las entidades deben escribirse con el formato: `nombre (tipo)`
- El formato para la entidad es una entidad y un solo par de paréntesis
- Las entidades pueden repetirse entre encabezados
- Si no aplica, listar 'Ninguno'
- Nada de explicaciones, solo escribe en formato markdown:
## [encabezado]
- nombre (tipo)
""",

fr="""
Tâche : Spécifier à quoi se réfèrent les titres indéfinis.

Instructions:
- Sélectionner parmi les titres autorisés uniquement ceux qui sont indéfinis (c'est-à-dire peu clairs, généraux, ambigus)
- Sélectionner signifie sélectionner, ne pas modifier les titres sélectionnés, ne pas en ajouter de nouveaux
- Pour chaque titre sélectionné, lister un par un chaque entité à laquelle le titre fait référence
- Les entités doivent être écrites au format : `nom (type)`
- Le format d'entité est une entité avec une seule paire de parenthèses
- Les entités peuvent se répéter entre les titres
- Si non applicable, lister 'Aucun'
- Pas d'explications, juste la sortie au format markdown :
## [titre]
- nom (type)
""",

de="""
Aufgabe: Geben Sie an, worauf sich die unbestimmten Überschriften beziehen.

Anweisungen:
- Wählen Sie aus den erlaubten Überschriften nur diejenigen aus, die unbestimmt sind (d. h. unklar, allgemein, mehrdeutig)
- Auswählen bedeutet auswählen, die ausgewählten Überschriften nicht ändern, keine neuen hinzufügen
- Listen Sie für jede ausgewählte Überschrift einzeln jede Entität auf, auf die sich die Überschrift bezieht
- Entitäten müssen im Format geschrieben werden: `Name (Typ)`
- Das Entitätsformat ist eine Entität mit einem einzigen Satz Klammern
- Entitäten können sich über Überschriften hinweg wiederholen
- Falls nicht zutreffend, listen Sie 'Keine' auf
- Keine Erklärungen, nur die Ausgabe im Markdown-Format:
## [Überschrift]
- Name (Typ)
""",
), examples=[

Example(flow="chat", example_input=Translations(
en="""
# Headings
## User (current user)
## Husband (spouse of user)
## Everyone (people at the dinner)

# The headings refer to entities in the following text
I had dinner with my family last night. My husband cooked lasagna, which is my daughter's favorite dish. Our son brought his girlfriend, who we were meeting for the first time. She brought a lovely bottle of wine that everyone liked.
""",
es="""
# Encabezados
## Usuario/a (usuario/a actual)
## Marido (cónyuge del usuario/a)
## Todos (personas en la cena)

# Los encabezados se refieren a entidades en el siguiente texto
Anoche cené con mi familia. Mi marido cocinó lasaña, que es el plato favorito de mi hija. Nuestro hijo trajo a su novia, a quien conocíamos por primera vez. Ella trajo una botella de vino encantadora que a todos les gustó.
""",
fr="""
# Titres
## Utilisateur/trice (utilisateur/trice actuel/le)
## Mari (conjoint de l'utilisateur/trice)
## Tout le monde (personnes présentes au dîner)

# Les titres font référence aux entités dans le texte suivant
J'ai dîné avec ma famille hier soir. Mon mari a cuisiné des lasagnes, qui est le plat préféré de ma fille. Notre fils a amené sa petite amie, que nous rencontrions pour la première fois. Elle a apporté une jolie bouteille de vin que tout le monde a aimée.
""",
de="""
# Überschriften
## Benutzer/in (aktueller Benutzer/in)
## Ehemann (Ehepartner des Benutzers/in)
## Alle (Personen beim Abendessen)

# Die Überschriften beziehen sich auf Entitäten im folgenden Text
Ich habe gestern Abend mit meiner Familie zu Abend gegessen. Mein Mann hat Lasagne gekocht, das ist das Lieblingsgericht meiner Tochter. Unser Sohn hat seine Freundin mitgebracht, die wir zum ersten Mal trafen. Sie hat eine schöne Flasche Wein mitgebracht, die allen gefiel.
""",
), example_output=Translations(
en="""
## Everyone (people at the dinner)
- User (current user)
- Husband (spouse of user)
- Daughter (child of user)
- Son (child of user)
- Girlfriend (girlfriend of user's son)
""",
es="""
## Todos (personas en la cena)
- Usuario/a (usuario/a actual)
- Esposo (cónyuge del usuario/a)
- Hija (hija del usuario/a)
- Hijo (hijo del usuario/a)
- Novia (novia del hijo del usuario/a)
""",
fr="""
## Tout le monde (personnes au dîner)
- Utilisateur/trice (utilisateur/trice actuel/le)
- Mari (conjoint/e de l'utilisateur/trice)
- Fille (enfant de l'utilisateur/trice)
- Fils (enfant de l'utilisateur/trice)
- Petite amie (petite amie du fils de l'utilisateur/trice)
""",
de="""
## Alle (Personen beim Abendessen)
- Benutzer/in (aktueller Benutzer/in)
- Ehemann (Ehepartner/in des Benutzers/in)
- Tochter (Kind des Benutzers/in)
- Sohn (Kind des Benutzers/in)
- Freundin (Freundin des Sohnes des Benutzers/in)
""",
)),

Example(flow="file", example_input=Translations(
en="""
# Headings
## Cytoplasm (internal cell fluid)
## All of them (organelles)
## Mitochondria (energy-producing organelle)

# The headings refer to entities in the following text
Eukaryotic cells have organelles like mitochondria (energy), endoplasmic reticulum (synthesis), and a nucleus (genetic information) enclosed by a nuclear envelope, all of which are contained in the cytoplasm.
""",
es="""
# Encabezados
## Citoplasma (fluido celular interno)
## Todos ellos (orgánulos)
## Mitocondrias (orgánulo productor de energía)

# Los encabezados se refieren a entidades en el siguiente texto
Las células eucariotas tienen orgánulos como las mitocondrias (energía), el retículo endoplasmático (síntesis) y un núcleo (información genética) encerrado por una envoltura nuclear, todos los cuales están contenidos en el citoplasma.
""",
fr="""
# Titres
## Cytoplasme (fluide cellulaire interne)
## Tous (organites)
## Mitochondries (organite producteur d'énergie)

# Les titres font référence à des entités dans le texte suivant
Les cellules eucaryotes possèdent des organites comme les mitochondries (énergie), le réticulum endoplasmique (synthèse) et un noyau (information génétique) entouré d'une enveloppe nucléaire, tous contenus dans le cytoplasme.
""",
de="""
# Überschriften
## Zytoplasma (innere Zellflüssigkeit)
## Alle (Organellen)
## Mitochondrien (energieproduzierendes Organell)

# Die Überschriften beziehen sich auf Entitäten im folgenden Text
Eukaryotische Zellen besitzen Organellen wie Mitochondrien (Energie), endoplasmatisches Retikulum (Synthese) und einen Zellkern (genetische Information), der von einer Kernhülle umschlossen ist, die alle im Zytoplasma enthalten sind.
""",
), example_output=Translations(
en="""
## All of them (organelles)
- mitochondria (organelle for energy)
- endoplasmic reticulum (organelle for synthesis)
- nucleus (organelle for genetic information)
""",
es="""
## Todos ellos (orgánulos)
- mitocondria (orgánulo para la energía)
- retículo endoplasmático (orgánulo para la síntesis)
- núcleo (orgánulo para la información genética)
""",
fr="""
## Tous (organites)
- mitochondrie (organite pour l'énergie)
- réticulum endoplasmique (organite pour la synthèse)
- noyau (organite pour l'information génétique)
""",
de="""
## Alle (Organellen)
- Mitochondrium (Organell für Energie)
- Endoplasmatisches Retikulum (Organell für Synthese)
- Zellkern (Organell für genetische Information)
""",

)),
])

QUERY_NODES = Prompt(prompt_type="system", prompt=Translations(

en="""
Task: Create Topics, Facts and Entities

Instructions:
- Derive topics from the INPUT and list them under '# Topics'
- Extract concise facts, if the INPUT is a question create concise potential answers, use format `## [fact or potential answer]`
- For each `## [fact or potential answer]`, list each entity (i.e. subjects, objects, events, concepts, etc.) with format `name (type)`
- Use 'None' if there are no items to list
- Use names instead of pronouns
- No explanations, just output in markdown format:
# Topics
- [topic]
- [topic]

## [fact or potential answer]
- name (type)
- name (type)
""",

es="""
Tarea: Crear Temas, Hechos y Entidades

Instrucciones:
- Derivar temas del INPUT y listarlos bajo '# Temas'
- Extraer hechos concisos, si el INPUT es una pregunta crea respuestas potenciales concisas, usar el formato `## [hecho o respuesta potencial]`
- Para cada `## [hecho o respuesta potencial]`, listar cada entidad (es decir, sujetos, objetos, eventos, conceptos, etc.) con el formato `nombre (tipo)`
- Usar 'None' si no hay elementos para listar
- Usar nombres en lugar de pronombres
- Nada de explicaciones, solo escribe en formato markdown:
# Temas
- [tema]
- [tema]

## [hecho o respuesta potencial]
- nombre (tipo)
- nombre (tipo)
""",

fr="""
Tâche: Créer des Sujets, des Faits et des Entités

Instructions:
- Dériver les sujets de l'INPUT et les lister sous '# Sujets'
- Extraire des faits concis, si l'INPUT est une question créer des réponses potentielles concises, utiliser le format `## [fait ou réponse potentielle]`
- Pour chaque `## [fait ou réponse potentielle]`, lister chaque entité (c'est-à-dire sujets, objets, événements, concepts, etc.) avec le format `nom (type)`
- Utiliser 'Aucun' s'il n'y a pas d'éléments à lister
- Utiliser des noms au lieu de pronoms
- Pas d'explications, juste sortir au format markdown:
# Sujets
- [sujet]
- [sujet]

## [fait ou réponse potentielle]
- nom (type)
- nom (type)
""",

de="""
Aufgabe: Themen, Fakten und Entitäten erstellen

Anweisungen:
- Themen aus dem INPUT ableiten und sie unter '# Themen' auflisten
- Prägnante Fakten extrahieren, wenn der INPUT eine Frage ist, prägnante potenzielle Antworten erstellen, Format `## [Fakt oder potenzielle Antwort]` verwenden
- Für jede(n) `## [Fakt oder potenzielle Antwort]`, jede Entität (d.h. Subjekte, Objekte, Ereignisse, Konzepte, etc.) im Format `Name (Typ)` auflisten
- 'Keine' verwenden, wenn keine Elemente zum Auflisten vorhanden sind
- Namen anstelle von Pronomen verwenden
- Keine Erklärungen, nur im Markdown-Format ausgeben:
# Themen
- [Thema]
- [Thema]

## [Fakt oder potenzielle Antwort]
- Name (Typ)
- Name (Typ)
""",

), examples=[

Example(flow="chat", example_input=Translations(
en="""
I had dinner with my family last night, we met our son's girlfriend for the first time.
""",
es="""
Cené con mi familia anoche, conocimos a la novia de nuestro hijo por primera vez.
""",
fr="""
J'ai dîné avec ma famille hier soir, nous avons rencontré la petite amie de notre fils pour la première fois.
""",
de="""
Ich habe gestern Abend mit meiner Familie zu Abend gegessen, wir haben die Freundin unseres Sohnes zum ersten Mal getroffen.
""",
), example_output=Translations(
en="""
# Topics
- Family dinner
- Family relationships

## The user had dinner with the user's family
- User (person)
- Dinner (event last night)
- Family (user's relatives)

## The user met the user's son's girlfriend for the first time.
- User (person)
- son (user's son)
- girlfriend (son's girlfriend)
""",
es="""
# Temas
- Cena familiar
- Relaciones familiares

## El usuario/a cenó con la familia del usuario/a
- Usuario/a (persona)
- Cena (evento de anoche)
- Familia (parientes del usuario/a)

## El usuario/a conoció a la novia del hijo del usuario/a por primera vez.
- Usuario/a (persona)
- hijo (hijo del usuario/a)
- novia (novia del hijo)
""",
fr="""
# Sujets
- Dîner de famille
- Relations familiales

## L'utilisateur/trice a dîné avec sa famille
- Utilisateur/trice (personne)
- Dîner (événement d'hier soir)
- Famille (proches de l'utilisateur/trice)

## L'utilisateur/trice a rencontré la petite amie de son fils pour la première fois.
- Utilisateur/trice (personne)
- fils (fils de l'utilisateur/trice)
- petite amie (petite amie du fils)
""",
de="""
# Themen
- Familienessen
- Familienbeziehungen

## Der Benutzer/in hat mit seiner Familie zu Abend gegessen
- Benutzer/in (Person)
- Abendessen (Ereignis von gestern Abend)
- Familie (Verwandte des Benutzers/in)

## Der Benutzer/in hat die Freundin seines Sohnes zum ersten Mal getroffen.
- Benutzer/in (Person)
- Sohn (Sohn des Benutzers/in)
- Freundin (Freundin des Sohnes)
""",
)),

Example(flow="file", example_input=Translations(
en="""
What is the time complexity of a breadth-first search (BFS) algorithm on a graph with V vertices and E edges, and why?
""",
es="""
¿Cuál es la complejidad temporal de un algoritmo de búsqueda en anchura (BFS) en un grafo con V vértices y E aristas, y por qué?
""",
fr="""
Quelle est la complexité temporelle d'un algorithme de parcours en largeur (BFS) sur un graphe avec V sommets et E arêtes, et pourquoi ?
""",
de="""
Wie hoch ist die Zeitkomplexität eines Breitensuche-Algorithmus (BFS) auf einem Graphen mit V Knoten und E Kanten, und warum?
""",
), example_output=Translations(
en="""
# Topics
- Graph search algorithm
- Time complexity formula

## Time complexity is O(V + E)
- Breadth-first search (algorithm)
- O (time complexity)
- V (vertices)
- E (edges)

## Each vertex is enqueued and dequeued at most once
- vertex (node)
- queue (data structure)
""",
es="""
# Temas
- Algoritmo de búsqueda en grafos
- Fórmula de complejidad temporal

## La complejidad temporal es O(V + E)
- Búsqueda en anchura (algoritmo)
- O (complejidad temporal)
- V (vértices)
- E (aristas)

## Cada vértice se encola y desencola como máximo una vez
- vértice (nodo)
- cola (estructura de datos)
""",
fr="""
# Sujets
- Algorithme de recherche de graphe
- Formule de complexité temporelle

## La complexité temporelle est O(V + E)
- Recherche en largeur (algorithme)
- O (complexité temporelle)
- V (sommets)
- E (arêtes)

## Chaque sommet est enfilé et défilé au plus une fois
- sommet (nœud)
- file (structure de données)
""",
de="""
# Themen
- Graphsuchalgorithmus
- Zeitkomplexitätsformel

## Die Zeitkomplexität ist O(V + E)
- Breitensuche (Algorithmus)
- O (Zeitkomplexität)
- V (Knoten)
- E (Kanten)

## Jeder Scheitelpunkt wird höchstens einmal in die Warteschlange gestellt und daraus entfernt
- Scheitelpunkt (Knoten)
- Warteschlange (Datenstruktur)
""",
)),
])

RETRIEVAL_CHAT = Prompt(prompt_type="system", prompt=Translations(

en="""
You are a helpful assistant. Memories are your own recollections. Use only relevant memories to inform your responses.
""",

es="""
Eres un asistente útil. Los recuerdos son tus propias rememoraciones. Utiliza solo recuerdos relevantes para informar tus respuestas.
""",

fr="""
Vous êtes un assistant utile. Les souvenirs sont vos propres souvenirs. Utilisez uniquement les souvenirs pertinents pour éclairer vos réponses.
""",

de="""
Sie sind ein hilfreicher Assistent. Erinnerungen sind Ihre eigenen Erinnerungen. Verwenden Sie nur relevante Erinnerungen, um Ihre Antworten zu informieren.
""",

), examples=[])


RETRIEVAL_RAG = Prompt(prompt_type="user", prompt=Translations(

en="""
Task: Respond to the user query using the provided sources.

Instructions:
- Use only relevant sources to inform your responses
- Incorporate inline citations in brackets: [id]
- Citation ids must correspond to the ids in the source tags `<source id=*>`
- If uncertain, concisely ask the user to rephrase the question to see if you can get better sources
""",

es="""
Tarea: Responder a la consulta del usuario utilizando las fuentes proporcionadas.

Instrucciones:
- Utiliza solo fuentes relevantes para informar tus respuestas
- Incorpora citas en línea entre corchetes: [id]
- Los ids de las citas deben corresponder a los ids en las etiquetas de fuente `<source id=*>`
- Si no estás seguro, pide concisamente al usuario que reformule la pregunta para ver si puedes obtener mejores fuentes
""",

fr="""
Tâche : Répondre à la requête de l'utilisateur en utilisant les sources fournies.

Instructions :
- Utilisez uniquement les sources pertinentes pour éclairer vos réponses
- Intégrez des citations en ligne entre crochets : [id]
- Les identifiants de citation doivent correspondre aux identifiants dans les balises source `<source id=*>`
- En cas d'incertitude, demandez de manière concise à l'utilisateur de reformuler la question pour voir si vous pouvez obtenir de meilleures sources
""",

de="""
Aufgabe: Beantworten Sie die Benutzeranfrage anhand der bereitgestellten Quellen.

Anweisungen:
- Verwenden Sie nur relevante Quellen, um Ihre Antworten zu gestalten
- Fügen Sie Inline-Zitate in Klammern ein: [id]
- Die Zitat-IDs müssen mit den IDs in den Quell-Tags `<source id=*>` übereinstimmen
- Wenn Sie unsicher sind, bitten Sie den Benutzer prägnant, die Frage neu zu formulieren, um zu sehen, ob Sie bessere Quellen erhalten können
""",

), examples=[

Example(flow="file", example_input=Translations(
en="""
<query>
What were the findings of the study on kv cache?
</query>

<source id="1">
Instead of predicting one token at a time, DeepSeek employs MTP, allowing the model to predict multiple future tokens in a single step.
6</source>
<source id="2">
Red Hat's blog post on integrating DeepSeek models with vLLM 0.7.1 highlights that MLA offers up to 9.6x more memory capacity for key-value (KV) caches.
11</source>
""",
es="""
<query>
¿Cuáles fueron los hallazgos del estudio sobre la caché KV?
</query>

<source id="1">
En lugar de predecir un token a la vez, DeepSeek emplea MTP, lo que permite al modelo predecir múltiples tokens futuros en un solo paso.
6</source>
<source id="2">
La publicación del blog de Red Hat sobre la integración de los modelos DeepSeek con vLLM 0.7.1 destaca que MLA ofrece hasta 9,6 veces más capacidad de memoria para las cachés de clave-valor (KV).
11</source>
""",
fr="""
<query>
Quelles ont été les conclusions de l'étude sur le cache KV ?
</query>

<source id="1">
Au lieu de prédire un token à la fois, DeepSeek utilise MTP, permettant au modèle de prédire plusieurs tokens futurs en une seule étape.
6</source>
<source id="2">
L'article de blog de Red Hat sur l'intégration des modèles DeepSeek avec vLLM 0.7.1 souligne que MLA offre jusqu'à 9,6 fois plus de capacité mémoire pour les caches clé-valeur (KV).
11</source>
""",
de="""
<query>
Was waren die Ergebnisse der Studie zum KV-Cache?
</query>

<source id="1">
Anstatt ein Token nach dem anderen vorherzusagen, verwendet DeepSeek MTP, wodurch das Modell mehrere zukünftige Tokens in einem einzigen Schritt vorhersagen kann.
6</source>
<source id="2">
Der Blogbeitrag von Red Hat zur Integration von DeepSeek-Modellen mit vLLM 0.7.1 hebt hervor, dass MLA bis zu 9,6x mehr Speicherkapazität für Key-Value (KV)-Caches bietet.
11</source>
""",
), example_output=Translations(
en="""
According to Red Hat's blog post on integrating DeepSeek models, MLA offers up to 9.6x more memory capacity for key-value caches [2].
""",
es="""
Según la publicación del blog de Red Hat sobre la integración de modelos DeepSeek, MLA ofrece hasta 9,6 veces más capacidad de memoria para cachés de clave-valor [2].
""",
fr="""
Selon l'article de blog de Red Hat sur l'intégration des modèles DeepSeek, MLA offre jusqu'à 9,6 fois plus de capacité mémoire pour les caches clé-valeur [2].
""",
de="""
Laut dem Blogbeitrag von Red Hat zur Integration von DeepSeek-Modellen bietet MLA bis zu 9,6-mal mehr Speicherkapazität für Schlüssel-Wert-Caches [2].
""",
)),
])


### FORMAT ###

FORMAT = Prompt(prompt_type="system", prompt=Translations(

en="""
""",

es="""
""",

fr="""
""",

de="""
""",

), examples=[

Example(flow="chat", example_input=Translations(
en="""
""",
es="""
""",
fr="""
""",
de="""
""",
), example_output=Translations(
en="""
""",
es="""
""",
fr="""
""",
de="""
""",
)),

Example(flow="file", example_input=Translations(
en="""
""",
es="""
""",
fr="""
""",
de="""
""",
), example_output=Translations(
en="""
""",
es="""
""",
fr="""
""",
de="""
""",
)),
])



import asyncio
import aiofiles
from neo4j import (
    AsyncGraphDatabase,
    exceptions as neo4jExceptions,
    SummaryCounters,
)
from typing import Any, Optional, List, Dict
from pydantic_settings import BaseSettings, SettingsConfigDict
from ragchat.log import get_logger, abbrev
from ragchat.utils import retry, timeit
from ragchat.definitions import get_base_target, NodeType, BaseNode, Relation, Flow, Filters, BaseFilters

logger = get_logger(__name__)

class Neo4jSettings(BaseSettings):
    url: Optional[str] = None
    local_hosts: Optional[List[str]] = ["localhost", "host.docker.internal"]
    bolt_port: int = 7687
    user: str = None
    password: str = None
    pool_size: int = 100
    timeout: int = 30
    acq_timeout: int = 60
    retry_time: int = 30
    retries: int = 3

    model_config = SettingsConfigDict(case_sensitive=False, env_prefix="NEO4J_")

    async def initialize(self):
        """
        Attempts to connect to Neo4j using the provided URL or default local hosts.
        Sets the `url` attribute to the first successful connection URL.
        Raises ConnectionError if no connection can be established.
        """
        urls_to_check = set()
        if self.url:
            urls_to_check.add(self.url)

        for host in self.local_hosts:
            urls_to_check.add(f"bolt://{host}:{self.bolt_port}")

        connection_attempts = [self._attempt_connection(url) for url in urls_to_check]
        results = await asyncio.gather(*connection_attempts, return_exceptions=True)
        self.url = next((result for result in results if result), None)
        if not self.url:
            raise ConnectionError(f"Could not connect to Neo4j using any of the default hosts or the provided URL: {self.url}")

        logger.info(f"Connection established using {self.url.split('@')[-1]}")

    async def _attempt_connection(self, url):
        """Attempts to connect to Neo4j at the given URL and returns the URL if successful."""
        try:
            driver = AsyncGraphDatabase.driver(url, auth=(self.user, self.password))
            async with driver:
                await driver.verify_connectivity()
            await driver.close()
            return url
        except Exception:
            return None

class MemoryGraph:
    def __init__(self, settings: Optional[Neo4jSettings] = None):
        self.settings = settings or Neo4jSettings()

        self.rerank_clauses = {
            "chunk": 0.0,
            "topic": 0.0,
            "fact": 0.0,
            "entity": "1 - apoc.text.jaroWinklerDistance(raw, n.content)",
        }
        self.rerank_weights = {
            "chunk": 0.0,
            "topic": 0.00,
            "fact": 0.00,
            "entity": 0.5,
        }

        self.retry_on = [ # only retry on these errors
            neo4jExceptions.TransientError,
            neo4jExceptions.ServiceUnavailable,
            neo4jExceptions.DatabaseError,
        ]

    async def initialize(self):
        """Initializes the Neo4j database with necessary constraints and indexes."""
        await self.settings.initialize()
        self.graph = AsyncGraphDatabase.driver(
            self.settings.url,
            auth=(self.settings.user, self.settings.password),
            max_connection_pool_size=self.settings.pool_size,
            connection_timeout=self.settings.timeout,
            connection_acquisition_timeout=self.settings.acq_timeout,
            max_transaction_retry_time=self.settings.retry_time,
        )

        queries  = [
            f"create index idx_{flow}_{k} if not exists for (n:{flow}) on (n.{k});"
            for flow in Flow
            for k in BaseNode.indexed_keys(flow)
        ]

        queries += [
            f"create index idx_chunk_node_id if not exists for (n:chunk) on (n.node_id);"
        ]

        async with self.graph.session() as session:
            tasks = [session.run(q) for q in queries]
            await asyncio.gather(*tasks)
            logger.info("Neo4j initialized successfully.")

    def _get_filter_clause(
        self,
        filters: BaseFilters,
        node_letter: str = 'n',
    ) -> str:
        """
        Build a Neo4j filter clause string based on provided filters.
        """
        with_custom = f"with *, apoc.convert.fromJsonMap({node_letter}.custom) as custom"
        filter_clauses = [
            f"coalesce({node_letter}.{k}, custom.{k}) {condition.operator} $filters.{k}"
            for k, v in filters.std_conditions().items()
            for condition in v if k in BaseNode.indexed_keys()
        ]

        if not filter_clauses:
            raise ValueError(f"Missing filter.")

        filter_clause_str = """
            and """.join(filter_clauses)
        filter_clause_str = f"""
            {with_custom}
            where {filter_clause_str}
        """ if filter_clauses else ""

        return filter_clause_str

    def _get_cleanup_clause(
        self,
        node_type: NodeType,
        node_letter: str = 'n',
    ) -> str:
        """
        Generate a Cypher clause to clean up orphaned nodes after deletion.

        The cleanup clause identifies and deletes:
        1. First-degree adjacent nodes that were connected only to the nodes being deleted and are different node_type from the nodes being deleted.
        2. Second-degree adjacent nodes connected only to nodes being deleted (original and first degree), thus about to become dangling nodes.

        Args:
            element_type: The node_type of element being deleted (e.g., NodeType.CHUNK)
            node_letter: The variable name used for the nodes being deleted in the preceding MATCH clause.

        Returns:
            str: Cypher clause for cleanup operations.

        NOTE: This clause assumes it's inserted *after* a MATCH clause identifying the nodes
              to delete (bound to `node_letter`) and *before* the final DETACH DELETE
              of those original nodes.
        """
        label = node_type.value # Get the string label from NodeType
        n = node_letter # alias for node_letter for readability
        r = f"""
            // Collect the original nodes identified for deletion
            with collect({n}) as x_nodes                                    // nodes being deleted
            with x_nodes as {n}_nodes, x_nodes                              // keep a pointer to the original nodes

            // 1. From first-degree adjacent nodes
            optional match (x)
            where not x:{label}                                             // Is not the node_type being deleted
            and exists((x)--(:{label}))                                     // Has connection to the node_type being deleted
            and all(y in [(x)--(z:{label}) | z] where y in x_nodes)         // Their only {label} connections are to nodes being deleted
            with {n}_nodes, x_nodes, x where x is not null                  // filter out nulls
            with {n}_nodes, x_nodes, collect(distinct x) as _nodes
            with {n}_nodes, x_nodes + _nodes as x_nodes                     // combine nodes being deleted

            // 2. From first/second-degree adjacent nodes
            optional match (x)
            where not x:{label}                                             // Is not the node_type being deleted
            and not any(y in [(x)--(z) | z] where not y in x_nodes)         // Connected only to nodes being deleted
            with {n}_nodes, x_nodes, x where x is not null                  // filter out nulls
            with {n}_nodes, x_nodes, collect(distinct x) as _nodes
            with {n}_nodes, x_nodes + _nodes as x_nodes                     // combine nodes being deleted

            // 3. delete cleanup nodes and unwind original nodes
            with {n}_nodes, x_nodes
            foreach (x in apoc.coll.subtract(x_nodes, {n}_nodes) | detach delete x)

            // 4. Unwind original nodes
            with {n}_nodes unwind {n}_nodes as {n}
        """

        return r

    def _get_projection(
            self,
            return_fields: set[str],
            node_type: NodeType = NodeType.CHUNK,
            as_map_str: bool = True,
            node_letter: str = "n",
            include: Optional[set[str]] = {"node_type", "flow"},
            exclude: Optional[set[str]] = {"embedding", "similarity"},
        ) -> dict | str:
        """
        Generates a Cypher projection string or dictionary for returning node properties.

        Args:
            return_fields: Set of fields to include in the projection. If None, uses default indexed keys for the node type.
            node_type: The type of node being projected (used for default keys).
            as_map_str: If True, returns a Cypher map string (e.g., "n{field1, field2}"). If False, returns a dictionary.
            node_letter: The variable name used for the node in the Cypher query.
            include: Set of fields to always include, even if not in return_fields (unless explicitly excluded).
            exclude: Set of fields to always exclude.

        Returns:
            dict | str: The Cypher projection as a dictionary or string.
        """
        return_fields = return_fields or (BaseNode.node_keys(node_type))
        return_fields = set(return_fields) - (exclude or set()) | (include or set())

        r = {field: f"{node_letter}.{field}" for field in return_fields}
        r |= {k: v for k, v in {
            "custom":       f"case when {node_letter}.custom is not null then apoc.convert.fromJsonMap({node_letter}.custom) else null end",
            "similarity":   f"similarity",
        }.items() if k in return_fields}

        if as_map_str:
            r = [f"{k}: {v}" for k, v in r.items()]
            r = node_letter + "{" + ", ".join(r) + "}"

        return r

    @retry()
    async def upsert_relations(
        self,
        relations: List[Relation],
        threshold: float,
        rr_threshold: float,
    ) -> List[Relation]:
        """
        Adds or updates relations and their associated nodes in the graph.

        Merges nodes if they already exist based on indexed keys or vector similarity
        (for new nodes without IDs). Sets properties and creates/merges relationships.

        Args:
            relations: A list of Relation objects to upsert.
            threshold: Vector similarity threshold for finding existing nodes.
            rr_threshold: Reranked similarity threshold for finding existing nodes.

        Returns:
            List[Relation]: The list of relations with updated node_ids for newly created nodes.
        """

        if not relations:
            return []

        if any(not (rel.chunk.model_fields_set & rel.chunk.indexed_keys(rel.chunk.flow)) for rel in relations):
            raise ValueError(f"Missing indexed field.")

        n_results = 0
        node_ids = {}
        async with self.graph.session() as session:
            for r in relations:
                # --- find existing nodes and their ids ---
                new_nodes = [n for n in r.to_list() if not n.node_id]
                # upsert ops use metadata as filters
                filters_target = get_base_target(BaseFilters, r.chunk.flow)
                filters = filters_target(**{k: getattr(r.chunk, k) for k in BaseNode.search_space() if getattr(r.chunk, k) is not None})
                if new_nodes:
                    # get ids if they already exist
                    for n in new_nodes:
                        n.node_id = node_ids.get(n._hash)

                    # re-filter, only nodes that don't have a node_id
                    new_nodes = [n for n in new_nodes if not n.node_id]

                if new_nodes:
                    # very similar nodes will be merged
                    return_fields = {"node_id", "content"}
                    tasks = [self.vector_search_nodes(n.node_type, filters, n.content, n.embedding, threshold, rr_threshold=rr_threshold, limit=1, return_fields=return_fields) for n in new_nodes]
                    existing_nodes = await asyncio.gather(*tasks)

                    # assign ids
                    for i, n in enumerate(new_nodes):
                        if not existing_nodes[i]:
                            continue
                        if existing_nodes[i][0].node_type == NodeType.CHUNK and existing_nodes[i][0] != n:
                            continue # chunks have to be exact match
                        n.node_id = existing_nodes[i][0].node_id
                        node_ids[n._hash] = n.node_id

                    # re-filter, only nodes that don't have a node_id
                    new_nodes = [n for n in new_nodes if not n.node_id]

                logger.debug(f"upserting relation: {abbrev(r)}")
                parameters = r.to_id_nodes()
                parameters.update({"filters": filters.std_dict()})

                # --- build cypher ---
                nodes = [r.fact, r.chunk] + r.topics + r.entities
                match_create_clauses = []
                # match or create clauses
                for i, n in enumerate(nodes):
                    with_clause = f"with {', '.join([f'n{j}' for j in range(i)])}\n" if i > 0 else ""
                    if n.node_id:
                        match_create_clauses.append(f"{with_clause}match (n{i}:{n.node_type}:{n.flow}) where n{i}.node_id = $n{i}.node_id")
                    else:
                        match_create_clauses.append(f"{with_clause}create (n{i}:{n.node_type}:{n.flow})")

                # set properties clauses
                set_clauses = []
                for i, n in enumerate(nodes):
                    weird_fields = {
                        "node_id": f"n{i}.node_id = coalesce(n{i}.node_id, randomUUID())",
                        "created_at": f"n{i}.created_at = coalesce(n{i}.created_at, timestamp())",
                        "updated_at": f"n{i}.updated_at = timestamp()",
                        "custom": f"n{i}.custom = case when coalesce($n{i}.custom, n{i}.custom) is not null then apoc.convert.toJson(coalesce($n{i}.custom, n{i}.custom)) else null end",
                    }

                    s  = [f"n{i}.{k} = coalesce($n{i}.{k}, n{i}.{k})" for k in BaseNode.node_keys(n.node_type) if k not in weird_fields]
                    s += [v for v in weird_fields.values()]
                    set_str = ",\n".join(s)

                    set_clauses.append(f"""
                    set
                        {set_str}
                    """)

                # relation clauses
                # chunk-[fact]->fact
                # topic-[fact]->fact
                # entity-[fact]->fact
                relation_clauses = []
                for i, n in enumerate(nodes):
                    if i != 0:
                        relation_clauses.append(f"merge (n{i})-[:{NodeType.FACT}]->(n0)")

                # return clauses
                return_clauses = [f"n{i}.node_id" for i in range(len(nodes))]

                # Combine all parts
                cypher = "\n".join(match_create_clauses)
                cypher += "\n" + "\n".join(set_clauses)
                cypher += "\n" + "\n".join(relation_clauses)
                cypher += "\nreturn " + ", ".join(return_clauses)

                # Execute the Cypher query
                try:
                    result = await session.run(cypher, parameters=parameters)
                    data = await result.data()
                    if data:
                        for i, n in enumerate(nodes):
                            n.node_id = data[0][f"n{i}.node_id"]
                            node_ids[n._hash] = n.node_id

                    n_results += 1

                except Exception as e:
                    if not any(isinstance(e, e_type) for e_type in self.retry_on):
                        logger.exception(f"Query:\n{cypher},\nParameters: {abbrev(parameters)}")
                    raise

        logger.info(f"Upserted {n_results}/{len(relations)} relations" + (f": {abbrev(relations)}" if logger.level <= 10 else ""))
        return relations

    @retry()
    async def get_nodes_by_ids(
        self,
        node_ids: List[str],
        node_type: NodeType = NodeType.CHUNK,
        return_fields: set[str] = None,
    ) -> List[BaseNode]:
        """
        Retrieve multiple nodes by their IDs.

        Args:
            node_ids: List of node IDs to retrieve
            element_type: Type of element to retrieve (default: CHUNK)
            return_fields: Fields to return for each node.

        Returns:
            List[Node]: The retrieved nodes (empty list if none found)
        """
        if not node_ids:
            return []
        filters = BaseFilters(node_ids=node_ids)

        cypher = f"""
            match (n:{node_type})
            {self._get_filter_clause(filters, 'n')}

            return {self._get_projection(return_fields)}
        """

        try:
            parameters = {"filters": filters.std_dict()}

            async with self.graph.session() as session:
                result = await session.run(cypher, parameters=parameters)
                records = await result.data()

            if not records:
                logger.debug(f"No {node_type} nodes found with IDs {node_ids}")
                return []

            nodes = []
            for record in records:
                nodes.append(BaseNode(**record['n']))

            return nodes

        except Exception as e:
            if not any(isinstance(e, e_type) for e_type in self.retry_on):
                logger.exception(f"Query:\n{cypher},\nParameters: {abbrev(parameters)}")
            raise

    @retry()
    async def get_nodes(
        self,
        filters: Filters,
        node_type: NodeType = NodeType.CHUNK,
        return_fields: Optional[set[str]] = None,
        order_by: str = 'updated_at',
        order_direction: str = 'desc',
        limit: int = 100,
    ) -> List[BaseNode]:
        """
        Retrieve all nodes of a specific node_type with filtering and ordering.

        Args:
            filters: Filters to apply to the query.
            node_type: Type of element to retrieve (default: CHUNK)
            limit: Maximum number of nodes to return (default: None, returns all)
            order_by: Property to order results by (default: None)
            order_direction: Direction of ordering, "asc" or "desc" (default: "asc")
            return_fields: Fields to return for each node.

        Returns:
            List[Node]: The retrieved nodes (empty list if none found)
        """
        if not order_by in ['created_at','updated_at']:
            raise ValueError(f"order_by must be a valid field, order_by: {order_by}")
        if not order_direction in ['asc', 'desc']:
            raise ValueError(f"order_direction must be a valid direction, order_direction: {order_direction}")
        if not filters:
            raise ValueError(f"Missing filters.")

        # Add ordering if specified
        order_clause = ""
        if order_by:
            order_clause = f"order by n.{order_by} {order_direction}"

        # Add limit if specified
        limit_clause = ""
        if limit is not None:
            limit_clause = f"limit {limit}"

        cypher = f"""
            match (n:{filters._flow}:{node_type})
            {self._get_filter_clause(filters, 'n')}
            {order_clause}
            {limit_clause}
            return {self._get_projection(return_fields, node_type=node_type)}
        """

        try:
            parameters = {"filters": filters.std_dict()}

            async with self.graph.session() as session:
                result = await session.run(cypher, parameters=parameters)
                records = await result.data()

            if not records:
                logger.debug(f"No {node_type} nodes found with the specified criteria")
                return []

            nodes = []
            for record in records:
                nodes.append(BaseNode(**record['n']))

            return nodes

        except Exception as e:
            if not any(isinstance(e, e_type) for e_type in self.retry_on):
                logger.exception(f"Query:\n{cypher},\nParameters: {abbrev(parameters)}")
            raise

    @retry()
    async def get_updated_at(
        self,
        filters: Filters,
        return_fields: set[str] = None,
        order_by: Optional[str] = 'updated_at',
        order_direction: Optional[str] = 'desc',
        limit: Optional[int] = 100,
    ) -> List[Dict[str, Any]]:
        """
        Retrieve the latest 'updated_at' timestamp for nodes matching filters, grouped by return fields.

        Args:
            filters: Filters to apply to the query.
            return_fields: Fields to group by and return along with the max updated_at. Defaults to indexed keys.
            limit: Maximum number of results to return (default: 100)
            order_by: Property to order results by ('created_at' or 'updated_at', default: 'updated_at')
            order_direction: Direction of ordering, "asc" or "desc" (default: "desc")
        """
        if not filters:
            raise ValueError(f"Missing filters.")
        if order_by and not order_by in ['created_at','updated_at']:
            raise ValueError(f"order_by must be a valid field, order_by: {order_by}")
        if order_direction and not order_direction in ['asc', 'desc']:
            raise ValueError(f"order_direction must be a valid direction, order_direction: {order_direction}")

        order_clause = ""
        if order_by:
            order_clause = f"order by n.{order_by} {order_direction}"

        limit_clause = ""
        if limit is not None:
            limit_clause = f"limit {limit}"

        return_fields = return_fields or BaseNode.indexed_keys()
        projection =", ".join([f"{v} as {k}" for k, v in self._get_projection(return_fields, as_map_str=False, include=None).items()])

        cypher = f"""
            match (n:{filters._flow}:chunk)
            {self._get_filter_clause(filters, 'n')}
            return {projection}, max(n.updated_at) as updated_at
            {order_clause}
            {limit_clause}
        """

        try:
            parameters = {"filters": filters.std_dict()}

            async with self.graph.session() as session:
                result = await session.run(cypher, parameters=parameters)
                records = await result.data()

            if not records:
                logger.debug(f"No chunk nodes found with the specified criteria")
                return []

            return records

        except Exception as e:
            if not any(isinstance(e, e_type) for e_type in self.retry_on):
                logger.exception(f"Query:\n{cypher},\nParameters: {abbrev(parameters)}")
            raise

    @retry()
    async def update_nodes_by_ids(
        self,
        nodes: List[BaseNode],
        return_fields: set[str] = None,
    ) -> List[BaseNode]:
        """
        Update multiple nodes by their IDs, preserving existing properties when new values are null.

        Args:
            nodes: List[Node objects with updated properties
            return_fields: Fields to return for the updated nodes.

        Returns:
            List[Node]: The updated nodes (empty list if none found)
        """
        if not nodes:
            return []

        node_data = []
        for node in nodes:
            properties = node.model_dump(mode="json", exclude={"node_id", "node_type", "flow"})
            node_data.append({
                "node_id": node.node_id,
                "node_type": node.node_type,
                "flow": node.flow,
                "properties": {k: v for k, v in properties.items()}
            })

        cypher = f"""
            unwind $node_data as data
            match (n:data.flow:data.node_type)
            where n.node_id = data.node_id

            set n = apoc.map.merge(n, data.properties, {{
                // For each property in the update, use coalesce to keep existing value if new is null
                // k, v -> case when v is null then n[k] else v end
                k, v -> coalesce(v, n[k])
            }})

            return {self._get_projection(return_fields)}
        """

        try:
            parameters = {"node_data": node_data}

            async with self.graph.session() as session:
                result = await session.run(cypher, parameters=parameters)
                records = await result.data()

            if not records:
                logger.debug(f"No nodes found with the provided IDs")
                return []

            updated_nodes = []
            for record in records:
                updated_nodes.append(BaseNode(**record['n']))

            return updated_nodes

        except Exception as e:
            if not any(isinstance(e, e_type) for e_type in self.retry_on):
                logger.exception(f"Query:\n{cypher},\nParameters: {abbrev(parameters)}")
            raise

    @retry()
    async def delete_nodes_by_ids(
        self,
        node_ids: List[str],
        node_type: NodeType = NodeType.CHUNK,
        return_fields: set[str] = None,
        cleanup: bool = True,
    ) -> List[BaseNode]:
        """
        Delete multiple nodes by their IDs.

        Args:
            node_ids: List of node IDs to delete
            element_type: Type of element to delete (default: CHUNK)
            return_fields: Fields to return for the deleted nodes.
            cleanup: Whether to perform cleanup operations on orphaned nodes.

        Returns:
            List[BaseNode]: The deleted nodes (empty list if none found).
        """
        if not node_ids:
            return []
        filters = BaseFilters(node_ids=node_ids)
        projection = self._get_projection(return_fields, as_map_str=True)[1:]

        cypher = f"""
            match (n:{node_type})
            {self._get_filter_clause(filters, 'n')}
            {self._get_cleanup_clause(node_type, 'n') if cleanup else ""}
            with n, {projection} as r
            detach delete n
            return r
        """

        try:
            parameters = {"filters": filters.std_dict()}

            async with self.graph.session() as session:
                result = await session.run(cypher, parameters=parameters)
                records = await result.data()

            if not records:
                logger.debug(f"No {node_type} nodes found with the specified criteria")
                return []

            nodes = [BaseNode(**r['r']) for r in records]

            return nodes

        except Exception as e:
            if not any(isinstance(e, e_type) for e_type in self.retry_on):
                logger.exception(f"Query:\n{cypher},\nParameters: {abbrev(parameters)}")
            raise

    @retry()
    async def delete_nodes(
        self,
        filters: Filters,
        node_type: NodeType = NodeType.CHUNK,
        return_fields: set[str] = None,
        cleanup: bool = True,
    ) -> List[BaseNode]:
        """
        Delete nodes based on filters.

        Args:
            element_type: Type of element to delete (default: chunk)
            filters: Filters to apply to the query (required)
            return_fields: Fields to return for the deleted nodes.
            cleanup: Whether to run cleanup operations for orphaned nodes (default: True)

        Returns:
            List[BaseNode]: The deleted nodes (empty list if none found).
        """
        if not filters:
            raise ValueError(f"Missing filters.")

        projection = self._get_projection(return_fields, as_map_str=True)[1:]

        cypher = f"""
            match (n:{filters._flow}:{node_type})
            {self._get_filter_clause(filters, 'n')}
            {self._get_cleanup_clause(node_type, 'n') if cleanup else ""}
            with n, {projection} as r
            detach delete n
            return r
        """

        try:
            parameters = {"filters": filters.std_dict()}

            async with self.graph.session() as session:
                result = await session.run(cypher, parameters=parameters)
                records = await result.data()

            if not records:
                logger.debug(f"No {node_type} nodes found with the specified criteria")
                return []

            nodes = [BaseNode(**r['r']) for r in records]

            return nodes


        except Exception as e:
            if not any(isinstance(e, e_type) for e_type in self.retry_on):
                logger.exception(f"Query:\n{cypher},\nParameters: {abbrev(parameters)}")
            raise

    @retry()
    async def vector_search_nodes(
        self,
        node_type: NodeType,
        filters: Filters,
        raw_string: str,
        embeddings: List[float],
        threshold: float,
        order_by: str = "similarity",
        limit: int = 100,
        rr_threshold: Optional[float] = None,
        return_fields: set[str] = None,
    ) -> List[BaseNode]:
        """
        Find the best matching nodes based on embedding similarity and optional reranking.

        Args:
            node_type: The type of node to search for.
            filters: Filters to apply to the search.
            raw_string: The original string used to generate the embeddings (used for reranking).
            embeddings: The embedding vector to search with.
            threshold: The minimum base similarity score required.
            order_by: Property to order results by ('similarity', 'created_at', or 'updated_at').
            limit: Maximum number of nodes to return.
            rr_threshold: The minimum reranked similarity score required (defaults to threshold).
            return_fields: Fields to return for each node.

        Returns:
            List[BaseNode]: A list of matching nodes, ordered by the specified property.
        """
        if not raw_string:
            return []
        if not filters:
            raise ValueError(f"Missing filters.")
        if not order_by in ['similarity','created_at','updated_at']:
            raise ValueError(f"order_by must be a valid field, order_by: {order_by}")
        if not rr_threshold:
            rr_threshold = threshold

        filter_clause = self._get_filter_clause(filters, 'n') or "where true"
        order_by = f"n.{order_by}" if order_by != "similarity" else "similarity"
        cypher = f"""
        match (n:{filters._flow}:{node_type})
        {filter_clause}
        and n.embedding is not null

        with n, $raw_string as raw,
        vector.similarity.cosine(n.embedding, $embedding) as base_similarity
        where base_similarity >= $threshold

        with n,
        (1.0 - {self.rerank_weights[node_type.value]}) * base_similarity + {self.rerank_weights[node_type.value]} * ({self.rerank_clauses[node_type.value]}) as similarity
        where similarity >= $rr_threshold
        order by {order_by} desc
        limit $limit

        return {self._get_projection(return_fields, node_type=node_type, include={"flow", "node_type", "similarity"})}
        order by {order_by} asc
        """

        try:
            parameters = {
                "raw_string": raw_string,
                "embedding": embeddings,
                "filters": filters.std_dict(),
                "threshold": threshold,
                "rr_threshold": rr_threshold,
                "limit": limit
            }

            async with self.graph.session() as session:
                result = await session.run(cypher, parameters=parameters)
                records = await result.data()

            nodes = [BaseNode(**r['n']) for r in records]

            return nodes

        except Exception as e:
            if not any(isinstance(e, e_type) for e_type in self.retry_on):
                logger.exception(f"Query:\n{cypher},\nParameters: {abbrev(parameters)}")
            raise

    @retry()
    @timeit
    async def search_chunks_from_relations(
        self,
        relations: List[Relation],
        filters: Filters,
        threshold: float,
        limit: int = 32,
        char_limit: int = 12000, # ~ 3k tokens (ollama default max 4k)
        explain: bool = False,
        return_fields: set[str] = None,
    ) -> List[BaseNode]:
        """
        Searches for relevant chunk nodes connected to the nodes within the provided relations.

        Uses vector similarity search on the related nodes (Fact, Topic, Entity) and
        then finds the connected Chunk nodes. Results are weighted and reranked based
        on similarity and node type connections.

        Args:
            relations: A list of Relation objects containing the input nodes (Fact, Topic, Entity) to search from.
            filters: Filters to apply to the search (applied to both input nodes and resulting chunks).
            threshold: The minimum similarity score required for related nodes.
            limit: Maximum number of chunk nodes to return.
            char_limit: Maximum total character count for the returned chunks.
            explain: If True, includes explanation details in the returned nodes.
            return_fields: Fields to return for each chunk node.

        Returns:
            List[BaseNode]: A list of relevant chunk nodes, limited by count and character limit,
                            and sorted by custom index ('ichunk') and creation time.
        """
        if not relations:
            return []
        if not filters:
            raise ValueError(f"Missing filters.")

        match_clauses = {
            f"{NodeType.TOPIC}":    f"(c:{NodeType.CHUNK})-->(:{NodeType.FACT})<--(n)",
            f"{NodeType.FACT}":     f"(c:{NodeType.CHUNK})-->(n)",
            f"{NodeType.ENTITY}":   f"(c:{NodeType.CHUNK})-->(:{NodeType.FACT})<--(n)",
        }

        call_clauses = [
            f"""
            with input_node
            where input_node.node_type = '{k}'
            match (n:{filters._flow}:{k})
            {self._get_filter_clause(filters, 'n')}
            and n.embedding is not null

            with input_node, input_node.content as raw, n, vector.similarity.cosine(input_node.embedding, n.embedding) as base_similarity
            where $threshold <= base_similarity

            match {match_clauses[k]}
            {self._get_filter_clause(filters, 'n')}

            with input_node, n, c,
                (1.0 - {self.rerank_weights[k]}) * base_similarity + {self.rerank_weights[k]} * ({self.rerank_clauses[k]}) as similarity

            return c, similarity
            {', "("+tostring(round(similarity * 1000.0) / 1000.0)+"="+input_node.content+"<=>"+n.content+")" as explain order by similarity desc' if explain else ''}
            """
            for k in match_clauses.keys()
        ]
        include = {"node_type", "flow", "similarity"}
        if explain:
            include |= {"types", "explain"}
        projection = self._get_projection(return_fields, node_letter='c', include=include)

        cypher = f"""
            unwind $input_nodes as input_node
            call (input_node) {{
                {"union".join(call_clauses)}
            }}

            // get max similarity per input-chunk
            with input_node, c, max(similarity) as similarity
            {', collect(explain)[0] as explain order by input_node.node_type' if explain else ', false as explain'}
            where $threshold <= similarity

            // count chunks
            with collect({{i: input_node, c: c, s: similarity, e: explain}}) as _map, count(distinct c) as total_chunks
            unwind _map as map
            with total_chunks, map.i as input_node, collect(map) as _map, count(distinct map.c) as input_chunks
            with input_node, _map, 1.0 + log(toFloat(total_chunks+1)/toFloat(input_chunks+1)) as weight
            {', total_chunks, input_chunks' if explain else ''}
            unwind _map as map

            with input_node, map.c as c, map.s as similarity, weight
            {', "(log("+total_chunks+"/"+input_chunks+")="+tostring(round(weight * 1000.0) / 1000.0)+")"+" * "+map.e as explain' if explain else ''}

            // get weighted avg similarity per node_type
            with c, input_node.node_type as node_type, avg(similarity) as similarity, avg(weight) as weight
            {', collect(explain) as explain'if explain else ''}
            with c, node_type, similarity^(1.0/weight) as similarity
            {', {node_type: node_type+"("+tostring(round(similarity*1000.0) / 1000.0)+"^(1/"+tostring(round(weight * 1000.0) / 1000.0)+")="+tostring(round(similarity^(1.0/weight)*1000.0)/1000.0)+")", explain: explain} as explain' if explain else ''}

            // get weighted avg similarity per chunk
            with c, collect(node_type) as types, avg(similarity) as similarity
            {', collect(explain) as explain' if explain else ''}
            where size(types) = {len(match_clauses)}
            order by similarity desc
            limit $limit

            return {projection}
            {' order by c.similarity desc' if explain else ''}
        """

        try:
            id_chunks = {}

            async def _get_id_chunks(r: Relation) -> Dict[str, Any]:
                nonlocal id_chunks
                input_nodes = r.to_list()
                if any(not n.embedding for n in input_nodes):
                    raise ValueError(f"Missing embedding for nodes.")

                parameters = {
                    "input_nodes": [n.model_dump(mode='json') for n in input_nodes],
                    "filters": filters.std_dict(),
                    "threshold": threshold,
                    "limit": limit,
                }

                async with self.graph.session() as session:
                    result = await session.run(cypher, parameters=parameters)
                    records = await result.data()

                if explain:
                    r = [{'c': {k: v for k, v in r['c'].items() if k != 'raw'}} for r in records]
                    logger.debug(f"records: {r}")

                return {r['c']['node_id']: r['c'] for r in records}

            tasks = [_get_id_chunks(r) for r in relations]
            id_chunks = {}
            for d in await asyncio.gather(*tasks):
                id_chunks.update(d)

            if not id_chunks:
                logger.debug(f"Nothing found. Filters {filters}")
                return []

            chunks = [BaseNode(**c) for c in id_chunks.values()]
            chunks.sort(key=lambda x: x.similarity)
            chunks = self._limit_chunks(chunks, limit, char_limit)
            chunks.sort(key=lambda x: (x.custom.get("ichunk", 0), x.created_at))

        except Exception as e:
            if not any(isinstance(e, e_type) for e_type in self.retry_on):
                logger.exception(f"Query:\n{cypher},\nFilters {filters}")
            raise

        return chunks

    @retry()
    async def export(
        self,
        file_name: str,
        filters: Filters,
        order_by: str = 'updated_at',
        order_direction: str = 'asc',
        return_fields: set[str] = None,
    ) -> None:
        """
        Dumps all nodes matching filters to a file, streaming the results to avoid loading everything into memory.

        Args:
            file_name: The name of the file to write to.
            filters: Filters to apply to the query.
            order_by: Property to order results by (default: 'updated_at').
            order_direction: Direction of ordering, "asc" or "desc" (default: "asc").
            return_fields: Fields to return for each node.
        """
        if not filters:
            raise ValueError(f"Missing filters.")
        if not order_by in ['created_at','updated_at']:
            raise ValueError(f"order_by must be a valid field, order_by: {order_by}")
        if not order_direction in ['asc', 'desc']:
            raise ValueError(f"order_direction must be a valid direction, order_direction: {order_direction}")

        # Add ordering if specified
        order_clause = ""
        if order_by:
            order_clause = f"order by n.{order_by} {order_direction}"

        cypher = f"""
            match (n:{filters._flow})
            {self._get_filter_clause(filters, 'n')}
            {order_clause}
            return {self._get_projection(return_fields)}
        """

        try:
            parameters = {"filters": filters.std_dict()}

            async with self.graph.session() as session:
                result = await session.run(cypher, parameters=parameters)

                async with aiofiles.open(file_name, 'w') as f:
                    async for record in result:
                        node = BaseNode(**record['n'])
                        await f.write(node.model_dump_json() + '\n')  # Write each node as a JSON string on a new line

            logger.info(f"Successfully exported nodes to {file_name}")

        except Exception as e:
            if not any(isinstance(e, e_type) for e_type in self.retry_on):
                logger.exception(f"Query:\n{cypher},\nParameters: {abbrev(parameters)}")
            raise

    async def close(self):
        """Closes the Neo4j graph driver connection."""
        await self.graph.close()

    def _limit_chunks(self, chunks: list[BaseNode], limit: int, char_limit: int) -> list[BaseNode]:
        """
        Limits the number of chunks and the total character count of their content.

        Args:
            chunks: The list of chunk nodes to limit.
            limit: The maximum number of chunks to return.
            char_limit: The maximum total character count allowed across all returned chunks.

        Returns:
            list[BaseNode]: The limited list of chunk nodes.
        """
        chunks = chunks[:limit]
        chars = 0
        limited_chunks = []
        for c in chunks:
            chars += len(c.content)
            if char_limit < chars:
                break
            limited_chunks.append(c)

        return limited_chunks




import os
import tempfile
import aiofiles
import asyncio
import asyncpg
from asyncpg import Pool, Connection
from typing import Any, List, Optional, Dict
from pydantic_settings import BaseSettings, SettingsConfigDict
from ragchat.utils import get_logger
from ragchat.definitions import FileMetadata

logger = get_logger(__name__)

class PgSettings(BaseSettings):
    url: Optional[str] = None
    local_hosts: Optional[List[str]] = ["localhost", "host.docker.internal"]
    user: str
    password: str
    port: int = 5432
    dbname: str
    prefetch: int = 5

    model_config = SettingsConfigDict(case_sensitive=False, env_prefix="PG_")

    async def initialize(self):
        """Attempts to connect to the PostgreSQL database."""
        urls_to_check = set()
        if self.url:
            urls_to_check.add(self.url)
        
        for host in self.local_hosts:
            urls_to_check.add(f"postgresql://{self.user}:{self.password}@{host}:{self.port}/{self.dbname}")

        connection_attempts = [self._attempt_connection(url) for url in urls_to_check]
        results = await asyncio.gather(*connection_attempts, return_exceptions=True)
        self.url = next((result for result in results if result), None)

        if not self.url:
            error_message = "Could not connect to PostgreSQL using any of the default hosts or the provided URL."
            if urls_to_check:
                # Mask the password in the URLs for logging
                masked_urls = [url.split('@')[-1] for url in urls_to_check]
                error_message += f" Attempted connections to: {', '.join(masked_urls)}"
            raise ConnectionError(error_message)

        logger.info(f"Connection established using ...{self.url.split('@')[-1]}")

    async def _attempt_connection(self, url):
        """Attempts to connect to PostgreSQL at the given url and returns the url if successful."""
        try:
            conn = await asyncpg.connect(url)
            await conn.close()
            return url
        except Exception:
            return None
        
class Pg():
    def __init__(self, settings: Optional[PgSettings] = None):
        """
        Initialize the Pg database.

        Args:
            settings (PgSettings): Configuration object
        """
        self.settings = settings or PgSettings()
        self.pool: Pool = None

    async def initialize(self):
        """Initialize the connection pool and create tables if needed"""
        await self.settings.initialize()
        self.pool = await asyncpg.create_pool(self.settings.url)

    async def download_files(self, file_ids: List[str], exclude_nones: bool = True) -> List[Optional[FileMetadata]]:
        """
        Retrieves file content for given IDs, streams to temporary files,
        and returns a list of dictionaries or None for each ID.

        Args:
            file_ids: A list of file IDs (strings).

        Returns:
            A list where each element corresponds to an input file_id.
            If the file was found and had content, the element is a dictionary
            {'id': str, 'name': str, 'path': str}.
            If the file was not found or had no content, the element is None.
        """
        if not file_ids:
            return []

        temp_dir = tempfile.mkdtemp(prefix="file_stream_cursor_")
        logger.debug(f"Created temporary directory: {temp_dir}")

        query_contents = """
		with 
		folders as (
			select id as folder_id,
			json_array_elements_text(data->'file_ids') as id
			from knowledge
		)
        select
            id as file_id,
            folder_id,
            filename,
            data->>'content' as file_text
        from file inner join folders using(id)
        where id = ANY($1::text[])
        """
        fetched_data_map: Dict[str, FileMetadata] = {}
        conn: Connection = None
        try:
            conn = await self.pool.acquire()
            async with conn.transaction():
                async for record in conn.cursor(query_contents, file_ids, prefetch=self.settings.prefetch):
                    file_id = str(record["file_id"])
                    file_text = record.get("file_text")

                    if file_text is not None:
                        temp_filename = f"{file_id}.txt"
                        temp_path = os.path.join(temp_dir, temp_filename)
                        try:
                            async with aiofiles.open(temp_path, mode='w', encoding='utf-8') as f:
                                await f.write(file_text)
                            fetched_data_map[file_id] = FileMetadata(
                                path=temp_path,
                                folder_id=record.get("folder_id"),
                                file_id=file_id,
                                title=record.get("filename"),
                            )
                            logger.debug(f"Successfully wrote file {file_id} to {temp_path}")
                        except Exception as e:
                            logger.warning(f"Could not write file {file_id} to temp path {temp_path}: {e}")
                    else:
                         logger.debug(f"File {file_id} found but had no content.")
        except Exception as db_error:
            logger.error(f"Error fetching files from database using cursor: {db_error}")
            raise db_error
        finally:
            if conn:
                await self.pool.release(conn)

        result_list: List[Optional[FileMetadata]] = [fetched_data_map.get(file_id) for file_id in file_ids]
        if exclude_nones:
            result_list = [item for item in result_list if item is not None]
        return result_list
    
    async def close(self):
        """Close the connection pool"""
        if self.pool:
            await self.pool.close()


# /home/raul/gitprojects/ragchat/memory/log.py
import reprlib
import logging
import sys
from typing import Optional, Any, Set
from pydantic_settings import BaseSettings, SettingsConfigDict

# --- LogSettings class remains the same ---
class LogSettings(BaseSettings):
    level: int = logging.INFO
    format: str = "%(asctime)s - RagChat:%(levelname)s: %(filename)s:%(lineno)d - %(message)s"
    date_format: str = "%H:%M:%S"
    file_path: Optional[str] = None

    model_config = SettingsConfigDict(case_sensitive=False, env_prefix="LOG_")

# --- COLORS, LEVEL_COLORS, _repr, abbrev, ColoredFormatter remain the same ---
# ANSI color codes
COLORS = {
    "RESET": "\033[0m",
    "RED": "\033[91m",
    "GREEN": "\033[92m",
    "YELLOW": "\033[93m",
    "BLUE": "\033[94m",
    "MAGENTA": "\033[95m",
    "CYAN": "\033[96m",
    "WHITE": "\033[97m",
}

# Map log levels to colors
LEVEL_COLORS = {
    logging.DEBUG: COLORS["BLUE"],
    logging.INFO: COLORS["GREEN"],
    logging.WARNING: COLORS["YELLOW"],
    logging.ERROR: COLORS["RED"],
    logging.CRITICAL: COLORS["MAGENTA"],
}


# Configure reprlib for abbreviated representations
_repr = reprlib.Repr()
_repr.maxlist = 4  # Show at most 4 elements from lists
_repr.maxstring = 100  # Limit string length
_repr.maxdict = 4  # Show at most 4 key-value pairs from dictionaries
_repr.maxset = 4  # Show at most 4 elements from sets
_repr.maxtuple = 4  # Show at most 4 elements from tuples
_repr.maxfrozenset = 4  # Show at most 4 elements from frozensets
_repr.maxdeque = 4  # Show at most 4 elements from deques
_repr.maxarray = 4  # Show at most 4 elements from arrays


def abbrev(obj: Any) -> str:
    """Create an abbreviated string representation of an object."""
    return _repr.repr(obj)


class ColoredFormatter(logging.Formatter):
    """Custom formatter that adds colors to log messages based on their level."""

    def format(self, record: logging.LogRecord) -> str:
        # Get the original formatted message
        message = super().format(record)

        # Find the position of the level in the message
        level_str = record.levelname
        # Handle potential case where levelname isn't found directly
        try:
            timestamp_and_logger = message.split(level_str)[0]
            rest_of_message = message[len(timestamp_and_logger) + len(level_str):]
        except ValueError:
             # Fallback if levelname isn't exactly in the formatted string
             timestamp_and_logger = ""
             rest_of_message = message

        # Apply color to the timestamp, logger name, and level
        color = LEVEL_COLORS.get(record.levelno, COLORS["RESET"])
        colored_message = (
            f"{color}{timestamp_and_logger}{level_str}{COLORS['RESET']}"
            f"{rest_of_message}"
        )

        return colored_message

# --- Global set to track loggers created by get_logger ---
_managed_loggers: Set[logging.Logger] = set()

# --- Store initial settings format/date_format for handler creation ---
# Read initial settings once for default formats
_initial_settings = LogSettings()
_default_format = _initial_settings.format
_default_date_format = _initial_settings.date_format
_default_file_path = _initial_settings.file_path


def get_logger(name: str) -> logging.Logger:
    """
    Gets or creates a logger, adds basic handlers, sets propagate=False,
    and tracks it for later configuration.
    Does NOT set the final level here.
    """
    logger = logging.getLogger(name)

    # Prevent duplicate handlers if called multiple times for the same logger
    if logger in _managed_loggers:
        return logger

    # Set propagate=False early to prevent issues
    logger.propagate = False

    # Check if appropriate handlers already exist before adding
    has_console_handler = any(
        isinstance(h, logging.StreamHandler) and h.stream == sys.stdout
        for h in logger.handlers
    )
    if not has_console_handler:
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.setFormatter(ColoredFormatter(_default_format, _default_date_format))
        logger.addHandler(console_handler)

    # Use the file path read initially or potentially updated later
    log_file_path = _default_file_path # Or read from current settings if needed dynamically
    if log_file_path:
         has_file_handler = any(
             isinstance(h, logging.FileHandler) and h.baseFilename == log_file_path
             for h in logger.handlers
         )
         if not has_file_handler:
            file_handler = logging.FileHandler(log_file_path)
            file_handler.setFormatter(logging.Formatter(_default_format, _default_date_format))
            logger.addHandler(file_handler)

    # Track this logger
    _managed_loggers.add(logger)

    return logger

def configure_logging(level: Optional[int] = None, configure_root: bool = True) -> None:
    """
    Reads the final LogSettings (after env vars are loaded) and applies
    the configuration (level) to all managed loggers and their handlers.
    Should be called by the main application after loading the environment.

    Args:
        level: An optional logging level to override the level from settings/env vars.
    """
    settings = LogSettings() # Reads current settings, potentially influenced by env vars

    # Determine the effective logging level
    effective_level = level if level is not None else settings.level

    # Update the global default file path in case it changed via env var
    global _default_file_path
    _default_file_path = settings.file_path

    logging.info(f"Reconfiguring {len(_managed_loggers)} managed loggers with level {effective_level} ({logging.getLevelName(effective_level)})")

    for logger in _managed_loggers:
        logger.setLevel(effective_level)
        for handler in logger.handlers:
            is_our_console = isinstance(handler, logging.StreamHandler) and handler.stream == sys.stdout
            is_our_file = isinstance(handler, logging.FileHandler) and settings.file_path and handler.baseFilename == settings.file_path

            if is_our_console or is_our_file:
                # Set the level (always do this)
                handler.setLevel(effective_level)

                # --- Update Formatter via env vars ---
                new_formatter = None
                if is_our_console:
                    new_formatter = ColoredFormatter(settings.format, settings.date_format)
                elif is_our_file:
                    new_formatter = logging.Formatter(settings.format, settings.date_format)

                if new_formatter:
                    handler.setFormatter(new_formatter)

    if configure_root:
        logging.getLogger().setLevel(effective_level)
